# -*- coding: utf-8 -*-
"""Utility functions.
"""
import numpy as np
import numba as nb

from sigpy import backend, config


__all__ = ['asscalar', 'prod', 'vec', 'split', 'rss', 'resize',
           'flip', 'circshift', 'downsample', 'upsample', 'dirac', 'randn',
           'triang', 'hanning', 'monte_carlo_sure', 'axpy', 'xpay',
           'ShuffledNumbers']


def _normalize_axes(axes, ndim):
    if axes is None:
        return tuple(range(ndim))
    else:
        return tuple(a % ndim for a in sorted(axes))


def _normalize_shape(shape):
    if isinstance(shape, int):
        return (shape, )
    else:
        return tuple(shape)


def _expand_shapes(*shapes):

    shapes = [list(shape) for shape in shapes]
    max_ndim = max(len(shape) for shape in shapes)
    shapes_exp = [[1] * (max_ndim - len(shape)) + shape
                  for shape in shapes]

    return tuple(shapes_exp)


def _check_same_dtype(*arrays):

    dtype = arrays[0].dtype
    for a in arrays:
        if a.dtype != dtype:
            raise TypeError(
                'inputs dtype mismatch, got {a_dtype}, and {dtype}.'.format(
                    a_dtype=a.dtype, dtype=dtype))


def asscalar(input):
    """Returns input array as scalar.

    Args:
        input (array): Input array

    Returns:
        scalar.

    """
    return np.asscalar(backend.to_device(input, backend.cpu_device))


def prod(shape):
    """Computes product of shape.

    Args:
        shape (tuple or list): shape.

    Returns:
        Product.

    """
    return np.prod(shape, dtype=np.long)


def vec(inputs):
    """Vectorize inputs.

    Args:
        shape (tuple or list): shape.

    Returns:
        array: Vectorized result.
    """
    device = backend.get_device(inputs[0])
    xp = device.xp

    with device:
        return xp.concatenate([i.ravel() for i in inputs])


def split(vec, oshapes):
    """Split input into specified output shapes.

    Args:
        oshapes (list of tuple of ints): Output shapes.

    Returns:
        list of arrays: Splitted outputs.
    """
    device = backend.get_device(vec)
    with device:
        outputs = []
        for oshape in oshapes:
            osize = prod(oshape)
            outputs.append(vec[:osize].reshape(oshape))
            vec = vec[osize:]

    return outputs


def rss(input, axes=(0, )):
    """Root sum of squares.

    Args:
        input (array): Input array.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Result.
    """

    device = backend.get_device(input)
    xp = device.xp

    with device:
        return xp.sum(xp.abs(input)**2, axis=axes)**0.5


def resize(input, oshape, ishift=None, oshift=None):
    """Resize with zero-padding or cropping.

    Args:
        input (array): Input array.
        oshape (tuple of ints): Output shape.
        ishift (None or tuple of ints): Input shift.
        oshift (None or tuple of ints): Output shift.

    Returns:
        array: Zero-padded or cropped result.
    """

    ishape_exp, oshape_exp = _expand_shapes(input.shape, oshape)

    if ishape_exp == oshape_exp:
        return input.reshape(oshape)

    if ishift is None:
        ishift = [max(i // 2 - o // 2, 0)
                  for i, o in zip(ishape_exp, oshape_exp)]

    if oshift is None:
        oshift = [max(o // 2 - i // 2, 0)
                  for i, o in zip(ishape_exp, oshape_exp)]

    copy_shape = [min(i - si, o - so) for i, si, o,
                  so in zip(ishape_exp, ishift, oshape_exp, oshift)]
    islice = tuple([slice(si, si + c) for si, c in zip(ishift, copy_shape)])
    oslice = tuple([slice(so, so + c) for so, c in zip(oshift, copy_shape)])

    device = backend.get_device(input)
    xp = device.xp
    with device:
        output = xp.zeros(oshape_exp, dtype=input.dtype)
        input = input.reshape(ishape_exp)
        output[oslice] = input[islice]

    return output.reshape(oshape)


def flip(input, axes=None):
    """Flip input.

    Args:
        input (array): Input array.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Flipped result.
    """

    if axes is None:
        axes = range(input.ndim)
    else:
        axes = _normalize_axes(axes, input.ndim)

    slc = []
    for d in range(input.ndim):
        if d in axes:
            slc.append(slice(None, None, -1))
        else:
            slc.append(slice(None))

    slc = tuple(slc)
    device = backend.get_device(input)
    with device:
        output = input[slc]

    return output


def circshift(input, shifts, axes=None):
    """Circular shift input.

    Args:
        input (array): Input array.
        shifts (tuple of ints): Shifts.
        axes (None or tuple of ints): Axes to perform operation.

    Returns:
        array: Result.
    """

    if axes is None:
        axes = range(input.ndim)

    assert(len(axes) == len(shifts))
    device = backend.get_device(input)
    xp = device.xp

    with device:
        for axis, shift in zip(axes, shifts):
            input = xp.roll(input, shift, axis=axis)

        return input


def downsample(input, factors, shift=None):
    """Downsample input.

    Args:
        input (array): Input array.
        factors (tuple of ints): Downsampling factors.
        shifts (None or tuple of ints): Shifts.

    Returns:
        array: Result.
    """

    if shift is None:
        shift = [0] * len(factors)

    slc = tuple(slice(s, None, f) for s, f in zip(shift, factors))

    device = backend.get_device(input)
    with device:
        return input[slc]


def upsample(input, oshape, factors, shift=None):
    """Upsample input.

    Args:
        input (array): Input array.
        factors (tuple of ints): Upsampling factors.
        shifts (None or tuple of ints): Shifts.

    Returns:
        array: Result.
    """

    if shift is None:
        shift = [0] * len(factors)

    slc = tuple(slice(s, None, f) for s, f in zip(shift, factors))

    device = backend.get_device(input)
    xp = device.xp
    with device:
        output = xp.zeros(oshape, dtype=input.dtype)
        output[slc] = input

    return output


def dirac(shape, dtype=np.float, device=backend.cpu_device):
    """Create Dirac delta.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: Dirac delta array.
    """

    device = backend.Device(device)
    xp = device.xp

    with device:
        return resize(xp.ones([1], dtype=dtype), shape)


def randn(shape, scale=1, dtype=np.float, device=backend.cpu_device):
    """Create random Gaussian array.

    Args:
        shape (tuple of ints): Output shape.
        scale (float): Standard deviation.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: Random Gaussian array.

    """
    device = backend.Device(device)
    xp = device.xp

    with device:
        if np.issubdtype(dtype, np.complexfloating):
            output = (xp.random.normal(
                size=shape, scale=scale / 2**0.5) * 1j).astype(dtype)
            output += xp.random.normal(size=shape, scale=scale / 2**0.5)
            return output
        else:
            return xp.random.normal(size=shape, scale=scale).astype(dtype)


def triang(shape, dtype=np.float, device=backend.cpu_device):
    """Create multi-dimensional triangular window.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: triangular filter.

    """
    device = backend.Device(device)
    xp = device.xp
    shape = _normalize_shape(shape)
    with device:
        window = xp.ones(shape, dtype=dtype)
        for n, i in enumerate(shape[::-1]):
            x = xp.arange(i, dtype=dtype)
            w = 1 - xp.abs(x - i // 2 + ((i + 1) % 2) / 2) / ((i + 1) // 2)
            window *= w.reshape([i] + [1] * n)

    return window


def hanning(shape, dtype=np.float, device=backend.cpu_device):
    """Create multi-dimensional hanning window.

    Args:
        shape (tuple of ints): Output shape.
        dtype (Dtype): Output data-type.
        device (Device): Output device.

    Returns:
        array: hanning filter.

    """
    device = backend.Device(device)
    xp = device.xp
    shape = _normalize_shape(shape)
    with device:
        window = xp.ones(shape, dtype=dtype)
        for n, i in enumerate(shape[::-1]):
            x = xp.arange(i, dtype=dtype)
            w = 0.5 - 0.5 * xp.cos(2 * np.pi * x / max(1, (i - (i % 2))))
            window *= w.reshape([i] + [1] * n)

    return window


def monte_carlo_sure(f, y, sigma, eps=1e-10):
    """Monte Carlo Stein Unbiased Risk Estimator (SURE).

    Monte carlo SURE assumes the observation y = x + e,
    where e is a white Gaussian array with standard deviation sigma.
    Monte carlo SURE provides an unbiased estimate of mean-squared error, ie:
    1 / n || f(y) - x ||_2^2

    Args:
        f (function): x -> f(x).
        y (array): observed measurement.
        sigma (float): noise standard deviation.

    Returns:
       float: SURE.

    References:
        Ramani, S., Blu, T. and Unser, M. 2008.
        Monte-Carlo Sure: A Black-Box Optimization of Regularization Parameters
        for General Denoising Algorithms. IEEE Transactions on Image Processing
        17, 9 (2008), 1540-1554.
    """
    device = backend.get_device(y)
    xp = device.xp

    n = y.size
    f_y = f(y)
    b = randn(y.shape, dtype=y.dtype, device=device)
    with device:
        divf_y = xp.real(xp.vdot(b, (f(y + eps * b) - f_y))) / eps
        sure = xp.mean(xp.abs(y - f_y)**2) - sigma**2 + \
            2 * sigma**2 * divf_y / n

    return sure


class ShuffledNumbers(object):
    """Produces shuffled numbers between given range.

    Args:
        Arguments to numpy.arange.

    """

    def __init__(self, *args):
        self.numbers = np.arange(*args)
        np.random.shuffle(self.numbers)
        self.i = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        ret = self.numbers[self.i]

        self.i += 1
        if self.i == len(self.numbers):
            np.random.shuffle(self.numbers)
            self.i = 0

        return ret


def axpy(y, a, x):
    """Compute y = a * x + y.

    Args:
        y (array): Output array.
        a (scalar): Input scalar.
        x (array): Input array.

    """
    device = backend.get_device(x)
    x = backend.to_device(x, device)
    a = backend.to_device(a, device)

    with device:
        if device == backend.cpu_device:
            _axpy(y, a, x, out=y)
        else:
            _axpy_cuda(a, x, y)


def xpay(y, a, x):
    """Compute y = x + a * y.

    Args:
        y (array): Output array.
        a (scalar): Input scalar.
        x (array): Input array.
    """

    device = backend.get_device(y)
    x = backend.to_device(x, device)
    a = backend.to_device(a, device)

    with device:
        if device == backend.cpu_device:
            _xpay(y, a, x, out=y)
        else:
            _xpay_cuda(a, x, y)


@nb.vectorize(nopython=True, cache=True)  # pragma: no cover
def _axpy(y, a, x):
    return a * x + y


@nb.vectorize(nopython=True, cache=True)  # pragma: no cover
def _xpay(y, a, x):
    return x + a * y


if config.cupy_enabled:  # pragma: no cover
    import cupy as cp

    _axpy_cuda = cp.ElementwiseKernel(
        'S a, T x',
        'T y',
        """
        y += (T) a * x;
        """,
        name='axpy')

    _xpay_cuda = cp.ElementwiseKernel(
        'S a, T x',
        'T y',
        """
        y = x + (T) a * y;
        """,
        name='axpy')



# -*- coding: utf-8 -*-
"""Convolution functions.

This module contains convolution functions that support multi-dimension,
and multi-channel.

"""
import numpy as np
from sigpy import backend, fourier, util, config


__all__ = ['convolve', 'convolve_adjoint_input', 'convolve_adjoint_filter']


def convolve(x, W, input_multi_channel=False,
             output_multi_channel=False, mode='full'):
    """Convolution that supports multi-dimensional and multi-channel inputs.

    Args:
        x (array): input array with shape batch_shape + input_shape
            if input_multi_channel=False.
            Otherwise with shape batch_shape + [input_channel] + input_shape.
        W (array): filter array of shape filter_shape,
            [input_channel] + filter_shape, [output_channel] + filter_shape,
            or [output_channel, input_channel] + filter_shape.
        input_multi_channel (bool): Specify if input has multiple channels.
        output_multi_channel (bool): Specify if output has multiple channels.
        mode (str): {'full', 'valid'}.

    Returns:
        array: output array with shape batch_shape + output_shape,
            if output_multi_channel=False.
            Otherwise with shape batch_shape + [output_channel] + output_shape
            if output_multi_channel=True.
            output_shape = input_shape + filter_shape - 1 if mode='full',
            output_shape = input_shape - filter_shape + 1 if mode='valid'.

    """
    ndim, input_shape, filter_shape, batch_shape, batch_size, \
        input_channel, output_channel = _get_convolve_params(
            x, W, input_multi_channel, output_multi_channel)

    device = backend.get_device(x)
    with device:
        x = x.reshape((batch_size, input_channel) + input_shape)
        W = W.reshape((output_channel, input_channel) + filter_shape)

    if (device != backend.cpu_device
        and config.cudnn_enabled):  # pragma: no cover
        y = _cudnn_convolve(x, W, mode=mode)
    else:
        y = _fft_convolve(x, W, mode=mode)

    output_shape = y.shape[-ndim:]
    with device:
        if output_multi_channel:
            return y.reshape(batch_shape + (output_channel, ) + output_shape)
        else:
            return y.reshape(batch_shape + output_shape)


def convolve_adjoint_input(W, y, input_multi_channel=False,
                           output_multi_channel=False, mode='full'):
    """Convolution adjoint.

    Args:
        W (array): filter array of shape filter_shape,
            [input_channel] + filter_shape, [output_channel] + filter_shape,
            or [output_channel, input_channel] + filter_shape.
        y (array): output array with shape batch_shape + output_shape
            if output_multi_channel=False.
            Otherwise with shape batch_shape + [output_channel] + input_shape.
        input_multi_channel (bool): Specify if input has multiple channels.
        output_multi_channel (bool): Specify if output has multiple channels.
        mode (str): {'full', 'valid'}.

    Returns:
        array: output array with shape batch_shape + output_shape,
            if output_multi_channel=False.
            Otherwise with shape batch_shape + [output_channel] + output_shape
            if output_multi_channel=True.
            output_shape = input_shape + filter_shape - 1 if mode='full',
            output_shape = input_shape - filter_shape + 1 if mode='valid'.

    """
    ndim, output_shape, filter_shape, batch_shape, batch_size, \
        input_channel, output_channel = _get_convolve_adjoint_input_params(
            W, y, input_multi_channel, output_multi_channel)

    device = backend.get_device(y)
    with device:
        y = y.reshape((batch_size, output_channel) + output_shape)
        W = W.reshape((output_channel, input_channel) + filter_shape)

    if (device != backend.cpu_device
        and config.cudnn_enabled):  # pragma: no cover
        x = _cudnn_convolve_adjoint_input(W, y, mode=mode)
    else:
        x = _fft_convolve_adjoint_input(W, y, mode=mode)

    input_shape = x.shape[-ndim:]
    with device:
        if input_multi_channel:
            return x.reshape(batch_shape + (input_channel, ) + input_shape)
        else:
            return x.reshape(batch_shape + input_shape)


def convolve_adjoint_filter(x, y, ndim, input_multi_channel=False,
                            output_multi_channel=False, mode='full'):
    """Convolution adjoint.

    Args:
        x (array): input array with shape batch_shape + input_shape
            if input_multi_channel=False.
            Otherwise with shape batch_shape + [input_channel] + input_shape.
        y (array): output array with shape batch_shape + output_shape
            if output_multi_channel=False.
            Otherwise with shape batch_shape + [output_channel] + input_shape.
        ndim (int): number of dimensions.
        input_multi_channel (bool): Specify if input has multiple channels.
        output_multi_channel (bool): Specify if output has multiple channels.
        mode (str): {'full', 'valid'}.

    Returns:
        array: filter array of shape filter_shape,
            [input_channel] + filter_shape, [output_channel] + filter_shape,
            or [output_channel, input_channel] + filter_shape.

    """
    input_shape, output_shape, batch_shape, batch_size, \
        input_channel, output_channel = _get_convolve_adjoint_filter_params(
            x, y, ndim, input_multi_channel, output_multi_channel)

    device = backend.get_device(x)
    with device:
        x = x.reshape((batch_size, input_channel) + input_shape)
        y = y.reshape((batch_size, output_channel) + output_shape)

    if (device != backend.cpu_device
        and config.cudnn_enabled):  # pragma: no cover
        W = _cudnn_convolve_adjoint_filter(x, y, mode=mode)
    else:
        W = _fft_convolve_adjoint_filter(x, y, mode=mode)

    with device:
        W_shape = W.shape[-ndim:]
        if input_multi_channel:
            W_shape = (input_channel, ) + W_shape

        if output_multi_channel:
            W_shape = (output_channel, ) + W_shape

        return W.reshape(W_shape)


def _get_convolve_params(x, W, input_multi_channel, output_multi_channel):
    ndim = W.ndim - input_multi_channel - output_multi_channel
    input_shape = x.shape[-ndim:]
    filter_shape = W.shape[-ndim:]
    batch_shape = x.shape[:-ndim - input_multi_channel]
    batch_size = util.prod(batch_shape)

    if x.dtype != W.dtype:
        raise TypeError(
            'x and W must have the same dtype, got {} and {}.'.format(
                x.dtype, W.dtype))

    if backend.get_device(x) != backend.get_device(W):
        raise TypeError(
            'x and W must be on the same device, got {} and {}.'.format(
                backend.get_device(x), backend.get_device(W)))

    if input_multi_channel:
        input_channel = x.shape[-ndim - 1]
    else:
        input_channel = 1

    if output_multi_channel:
        output_channel = W.shape[
            -ndim - input_multi_channel - output_multi_channel]
    else:
        output_channel = 1

    return ndim, input_shape, filter_shape, batch_shape, \
        batch_size, input_channel, output_channel


def _get_convolve_adjoint_input_params(
        W, y, input_multi_channel, output_multi_channel):
    ndim = W.ndim - input_multi_channel - output_multi_channel
    output_shape = y.shape[-ndim:]
    filter_shape = W.shape[-ndim:]
    batch_shape = y.shape[:-ndim - output_multi_channel]
    batch_size = util.prod(batch_shape)

    if y.dtype != W.dtype:
        raise TypeError(
            'y and W must have the same dtype, got {} and {}.'.format(
                y.dtype, W.dtype))

    if backend.get_device(y) != backend.get_device(W):
        raise TypeError(
            'y and W must be on the same device, got {} and {}.'.format(
                backend.get_device(y), backend.get_device(W)))

    if input_multi_channel:
        input_channel = W.shape[-ndim - 1]
    else:
        input_channel = 1

    if output_multi_channel:
        output_channel = y.shape[-ndim - 1]
    else:
        output_channel = 1

    return ndim, output_shape, filter_shape, batch_shape, \
        batch_size, input_channel, output_channel


def _get_convolve_adjoint_filter_params(
        x, y, ndim, input_multi_channel, output_multi_channel):
    output_shape = y.shape[-ndim:]
    input_shape = x.shape[-ndim:]
    batch_shape = y.shape[:-ndim - output_multi_channel]
    batch_size = util.prod(batch_shape)

    if x.dtype != y.dtype:
        raise TypeError(
            'x and y must have the same dtype, got {} and {}.'.format(
                x.dtype, y.dtype))

    if backend.get_device(y) != backend.get_device(x):
        raise TypeError(
            'y and x must be on the same device, got {} and {}.'.format(
                backend.get_device(y), backend.get_device(x)))

    if input_multi_channel:
        input_channel = x.shape[-ndim - 1]
    else:
        input_channel = 1

    if output_multi_channel:
        output_channel = y.shape[-ndim - 1]
    else:
        output_channel = 1

    return input_shape, output_shape, batch_shape, \
        batch_size, input_channel, output_channel


def _fft_convolve(x, W, mode='full'):
    ndim = x.ndim - 2
    batch_size = len(x)
    output_channel, input_channel = W.shape[:2]
    input_shape = x.shape[-ndim:]
    filter_shape = W.shape[-ndim:]
    if mode == 'full':
        output_shape = tuple(
            m + n - 1 for m, n in zip(input_shape, filter_shape))
        pad_shape = output_shape
    elif mode == 'valid':
        output_shape = tuple(
            m - n + 1 for m, n in zip(input_shape, filter_shape))
        pad_shape = input_shape

    dtype = x.dtype
    device = backend.get_device(x)
    xp = device.xp
    with device:
        x = x.reshape((batch_size, 1, input_channel) + input_shape)
        x_pad = util.resize(x, (batch_size, 1, input_channel) + pad_shape,
                            oshift=[0] * x.ndim)
        W_pad = util.resize(W, (output_channel, input_channel) + pad_shape,
                            oshift=[0] * W.ndim)

        if np.issubdtype(dtype, np.floating):
            x_fft = xp.fft.rfftn(x_pad, axes=range(-ndim, 0), norm='ortho')
            W_fft = xp.fft.rfftn(W_pad, axes=range(-ndim, 0), norm='ortho')
            y_fft = xp.sum(x_fft * W_fft, axis=-ndim - 1)
            y = xp.fft.irfftn(y_fft, pad_shape,
                              axes=range(-ndim, 0), norm='ortho').astype(dtype)
        else:
            x_fft = fourier.fft(x_pad, axes=range(-ndim, 0), center=False)
            W_fft = fourier.fft(W_pad, axes=range(-ndim, 0), center=False)
            y_fft = xp.sum(x_fft * W_fft, axis=-ndim - 1)
            y = fourier.ifft(y_fft, axes=range(-ndim, 0), center=False)

        if mode == 'full':
            shift = [0] * y.ndim
        elif mode == 'valid':
            shift = [0, 0] + [n - 1 for n in filter_shape]

        y = util.resize(y, (batch_size, output_channel) + output_shape,
                        ishift=shift)
        y *= util.prod(pad_shape)**0.5
        return y


def _fft_convolve_adjoint_input(W, y, mode='full'):
    ndim = y.ndim - 2
    batch_size = len(y)
    output_channel, input_channel = W.shape[:2]
    output_shape = y.shape[-ndim:]
    filter_shape = W.shape[-ndim:]
    if mode == 'full':
        input_shape = tuple(
            p - n + 1 for p, n in zip(output_shape, filter_shape))
        pad_shape = output_shape
    elif mode == 'valid':
        input_shape = tuple(
            p + n - 1 for p, n in zip(output_shape, filter_shape))
        pad_shape = input_shape

    dtype = y.dtype
    device = backend.get_device(y)
    xp = device.xp
    with device:
        y = y.reshape((batch_size, output_channel, 1) + output_shape)
        W = xp.conj(util.flip(W, axes=range(-ndim, 0)))

        y_pad = util.resize(y, (batch_size, output_channel, 1) + pad_shape,
                            oshift=[0] * y.ndim)
        W_pad = util.resize(W, (output_channel, input_channel) + pad_shape,
                            oshift=[0] * W.ndim)

        if np.issubdtype(dtype, np.floating):
            y_fft = xp.fft.rfftn(y_pad, axes=range(-ndim, 0), norm='ortho')
            W_fft = xp.fft.rfftn(W_pad, axes=range(-ndim, 0), norm='ortho')
            x_fft = xp.sum(y_fft * W_fft, axis=-ndim - 2)
            x = xp.fft.irfftn(x_fft, pad_shape,
                              axes=range(-ndim, 0), norm='ortho').astype(dtype)
        else:
            y_fft = fourier.fft(y_pad, axes=range(-ndim, 0), center=False)
            W_fft = fourier.fft(W_pad, axes=range(-ndim, 0), center=False)
            x_fft = xp.sum(y_fft * W_fft, axis=-ndim - 2)
            x = fourier.ifft(x_fft, axes=range(-ndim, 0), center=False)

        if mode == 'full':
            shift = [0, 0] + [n - 1 for n in filter_shape]
        elif mode == 'valid':
            shift = [0] * x.ndim

        x = util.resize(
            x, (batch_size, input_channel) + input_shape, ishift=shift)
        x *= util.prod(pad_shape)**0.5
        return x


def _fft_convolve_adjoint_filter(x, y, mode='full'):
    ndim = x.ndim - 2
    batch_size = len(x)
    output_channel = y.shape[1]
    input_channel = x.shape[1]
    output_shape = y.shape[-ndim:]
    input_shape = x.shape[-ndim:]
    if mode == 'full':
        filter_shape = tuple(
            p - m + 1 for m, p in zip(input_shape, output_shape))
        pad_shape = output_shape
    elif mode == 'valid':
        filter_shape = tuple(
            m - p + 1 for m, p in zip(input_shape, output_shape))
        pad_shape = input_shape

    dtype = x.dtype
    device = backend.get_device(x)
    xp = device.xp
    with device:
        x = xp.conj(util.flip(x, axes=range(-ndim, 0)))
        x = x.reshape((batch_size, 1, input_channel) + input_shape)
        y = y.reshape((batch_size, output_channel, 1) + output_shape)

        x_pad = util.resize(x, (batch_size, 1, input_channel) + pad_shape,
                            oshift=[0] * x.ndim)
        y_pad = util.resize(y, (batch_size, output_channel, 1) + pad_shape,
                            oshift=[0] * y.ndim)

        if np.issubdtype(dtype, np.floating):
            x_fft = xp.fft.rfftn(x_pad, axes=range(-ndim, 0), norm='ortho')
            y_fft = xp.fft.rfftn(y_pad, axes=range(-ndim, 0), norm='ortho')
            W_fft = xp.sum(x_fft * y_fft, axis=0)
            W = xp.fft.irfftn(W_fft, pad_shape,
                              axes=range(-ndim, 0), norm='ortho').astype(dtype)
        else:
            x_fft = fourier.fft(x_pad, axes=range(-ndim, 0), center=False)
            y_fft = fourier.fft(y_pad, axes=range(-ndim, 0), center=False)
            W_fft = xp.sum(x_fft * y_fft, axis=0)
            W = fourier.ifft(W_fft, axes=range(-ndim, 0), center=False)

        if mode == 'full':
            shift = [0, 0] + [m - 1 for m in input_shape]
        elif mode == 'valid':
            shift = [0, 0] + [p - 1 for p in output_shape]

        W = util.resize(
            W, (output_channel, input_channel) + filter_shape, ishift=shift)
        W *= util.prod(pad_shape)**0.5
        return W


if config.cudnn_enabled:  # pragma: no cover
    from cupy import cudnn

    def _cudnn_convolve(x, W, mode='full'):
        dtype = x.dtype
        device = backend.get_device(x)
        xp = device.xp
        if np.issubdtype(dtype, np.complexfloating):
            with device:
                xr = xp.real(x)
                xi = xp.imag(x)
                Wr = xp.real(W)
                Wi = xp.imag(W)

                # Concatenate real and imaginary to input/output channels
                x = xp.concatenate([xr, xi], axis=1)
                W = xp.concatenate([xp.concatenate([Wr, -Wi], axis=1),
                                    xp.concatenate([Wi, Wr], axis=1)], axis=0)

                y = _cudnn_convolve(x, W, mode=mode)

                # Convert back to complex
                y = y[:, :y.shape[1] // 2] + 1j * y[:, y.shape[1] // 2:]
                y = y.astype(dtype)

                return y

        ndim = x.ndim - 2
        batch_size = len(x)
        output_channel = W.shape[0]
        input_shape = x.shape[-ndim:]
        filter_shape = W.shape[-ndim:]
        strides = (1, ) * ndim
        dilations = (1, ) * ndim
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        if mode == 'full':
            output_shape = tuple(
                m + n - 1 for m, n in zip(input_shape, filter_shape))
            pads = tuple(n - 1 for n in W.shape[2:])
        elif mode == 'valid':
            output_shape = tuple(
                m - n + 1 for m, n in zip(input_shape, filter_shape))
            pads = (0, ) * ndim

        with device:
            y = xp.empty((batch_size, output_channel) + output_shape,
                         dtype=dtype)
            W = util.flip(W, axes=range(-ndim, 0))
            cudnn.convolution_forward(x, W, None, y,
                                      pads, strides, dilations, groups,
                                      auto_tune=auto_tune,
                                      tensor_core=tensor_core)

        return y

    def _cudnn_convolve_adjoint_input(W, y, mode='full'):
        dtype = y.dtype
        device = backend.get_device(y)
        xp = device.xp
        if np.issubdtype(dtype, np.complexfloating):
            with device:
                Wr = xp.real(W)
                Wi = xp.imag(W)
                yr = xp.real(y)
                yi = xp.imag(y)

                # Concatenate real and imaginary to input/output channels
                y = xp.concatenate([yr, yi], axis=1)
                W = xp.concatenate([xp.concatenate([Wr, -Wi], axis=1),
                                    xp.concatenate([Wi, Wr], axis=1)], axis=0)

                x = _cudnn_convolve_adjoint_input(W, y, mode=mode)

                # Convert back to complex
                x = x[:, :x.shape[1] // 2] + 1j * x[:, x.shape[1] // 2:]
                x = x.astype(dtype)

                return x

        ndim = y.ndim - 2
        batch_size = len(y)
        input_channel = W.shape[1]
        output_shape = y.shape[-ndim:]
        filter_shape = W.shape[-ndim:]
        strides = (1, ) * ndim
        dilations = (1, ) * ndim
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        deterministic = False
        if mode == 'full':
            input_shape = tuple(
                p - n + 1 for p, n in zip(output_shape, filter_shape))
            pads = tuple(n - 1 for n in W.shape[2:])
        elif mode == 'valid':
            input_shape = tuple(
                p + n - 1 for p, n in zip(output_shape, filter_shape))
            pads = (0, ) * ndim

        with device:
            x = xp.empty((batch_size, input_channel) + input_shape,
                         dtype=dtype)
            W = util.flip(W, axes=range(-ndim, 0))
            cudnn.convolution_backward_data(W, y, None, x,
                                            pads, strides, dilations, groups,
                                            deterministic=deterministic,
                                            auto_tune=auto_tune,
                                            tensor_core=tensor_core)

        return x

    def _cudnn_convolve_adjoint_filter(x, y, mode='full'):
        dtype = y.dtype
        device = backend.get_device(y)
        xp = device.xp
        if np.issubdtype(dtype, np.complexfloating):
            with device:
                xr = xp.real(x)
                xi = xp.imag(x)
                yr = xp.real(y)
                yi = xp.imag(y)

                # Concatenate real and imaginary to input/output channels
                x = xp.concatenate([xr, xi], axis=1)
                y = xp.concatenate([yr, yi], axis=1)

                W = _cudnn_convolve_adjoint_filter(x, y, mode=mode)

                # Convert back to complex
                Wr = W[:W.shape[0] // 2, :W.shape[1] // 2]
                Wr += W[W.shape[0] // 2:, W.shape[1] // 2:]
                Wi = W[W.shape[0] // 2:, :W.shape[1] // 2]
                Wi -= W[:W.shape[0] // 2, W.shape[1] // 2:]
                return (Wr + 1j * Wi).astype(dtype)

        ndim = y.ndim - 2
        input_channel = x.shape[1]
        output_channel = y.shape[1]
        input_shape = x.shape[-ndim:]
        output_shape = y.shape[-ndim:]
        strides = (1, ) * ndim
        dilations = (1, ) * ndim
        groups = 1
        auto_tune = True
        tensor_core = 'auto'
        deterministic = False
        if mode == 'full':
            filter_shape = tuple(
                p - m + 1 for m, p in zip(input_shape, output_shape))
            pads = tuple(n - 1 for n in filter_shape)
        elif mode == 'valid':
            filter_shape = tuple(
                m - p + 1 for m, p in zip(input_shape, output_shape))
            pads = (0, ) * ndim

        with device:
            W = xp.empty(
                (output_channel, input_channel) + filter_shape, dtype=dtype)
            cudnn.convolution_backward_filter(x, y, W,
                                              pads, strides, dilations, groups,
                                              deterministic=deterministic,
                                              auto_tune=auto_tune,
                                              tensor_core=tensor_core)
            W = util.flip(W, axes=range(-ndim, 0))

        return W



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import alg

if __name__ == '__main__':
    unittest.main()


class TestAlg(unittest.TestCase):
    def Ax_setup(self, n):
        A = np.eye(n) + 0.1 * np.ones([n, n])
        x = np.arange(n, dtype=np.float)
        return A, x

    def Ax_y_setup(self, n, lamda):
        A, x = self.Ax_setup(n)
        y = A @ x
        x_numpy = np.linalg.solve(
            A.T @ A + lamda * np.eye(n), A.T @ y)

        return A, x_numpy, y

    def test_PowerMethod(self):
        n = 5
        A, x = self.Ax_setup(n)
        x_hat = np.random.random([n, 1])
        alg_method = alg.PowerMethod(lambda x: A.T @ A @ x, x_hat)
        while(not alg_method.done()):
            alg_method.update()

        s_numpy = np.linalg.svd(A, compute_uv=False)[0]
        s_sigpy = np.linalg.norm(A @ x_hat)
        npt.assert_allclose(s_numpy, s_sigpy, atol=1e-3)

    def test_GradientMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Compute step-size
        lipschitz = np.linalg.svd(
            A.T @ A + lamda * np.eye(n), compute_uv=False)[0]
        alpha = 1.0 / lipschitz

        for beta in [1, 0.5]:
            for accelerate in [True, False]:
                for proxg in [None, lambda alpha, x: x / (1 + lamda * alpha)]:
                    with self.subTest(accelerate=accelerate,
                                      proxg=proxg,
                                      beta=beta):
                        x_sigpy = np.zeros([n])

                        def gradf(x):
                            gradf_x = A.T @ (A @ x - y)
                            if proxg is None:
                                gradf_x += lamda * x

                            return gradf_x

                        if beta == 1:
                            alpha = 1 / lipschitz
                            f = None
                        else:
                            alpha = 1

                            def f(x):
                                f_x = 1 / 2 * np.linalg.norm(A @ x - y)**2
                                if proxg is None:
                                    f_x += lamda / 2 * np.linalg.norm(x)**2

                                return f_x

                        alg_method = alg.GradientMethod(
                            gradf,
                            x_sigpy,
                            alpha,
                            beta=beta,
                            f=f,
                            accelerate=accelerate,
                            proxg=proxg,
                            max_iter=1000)

                        while(not alg_method.done()):
                            alg_method.update()

                        npt.assert_allclose(x_sigpy, x_numpy)

    def test_ConjugateGradient(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)
        x = np.zeros([n])
        alg_method = alg.ConjugateGradient(
            lambda x: A.T @ A @ x + lamda * x,
            A.T @ y, x, max_iter=1000)
        while(not alg_method.done()):
            alg_method.update()

        npt.assert_allclose(x, x_numpy)

    def test_PrimalDualHybridGradient(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Compute step-size
        lipschitz = np.linalg.svd(np.matmul(A.T, A), compute_uv=False)[0]
        tau = 1.0 / lipschitz
        sigma = 1.0

        x = np.zeros([n])
        u = np.zeros([n])
        alg_method = alg.PrimalDualHybridGradient(
            lambda alpha, u: (u - alpha * y) / (1 + alpha),
            lambda alpha, x: x / (1 + lamda * alpha),
            lambda x: A @ x,
            lambda x: A.T @ x,
            x, u, tau, sigma, max_iter=1000)
        while(not alg_method.done()):
            alg_method.update()

        npt.assert_allclose(x, x_numpy)

    def test_AugmentedLagrangianMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        # Solve 1 / 2 \| A x - y \|_2^2 + lamda * \| z \|_2^2 s.t. x = z
        mu = 1
        x_z = np.zeros([2 * n])
        u = np.zeros([n])

        def min_lagrangian(x_z, u, mu):
            x = x_z[:n]
            z = x_z[n:]
            x[:] = np.linalg.solve(
                A.T @ A + mu * np.eye(n), A.T @ y + u + mu * z)
            z[:] = (mu * x - u) / (mu + lamda)

        def constraints(x_z):
            x = x_z[:n]
            z = x_z[n:]
            return x - z

        alg_method = alg.AugmentedLagrangianMethod(
            min_lagrangian, constraints, x_z, u, mu)
        while(not alg_method.done()):
            alg_method.update()

        x = x_z[:n]
        npt.assert_allclose(x, x_numpy)

    def test_NewtonsMethod(self):
        n = 5
        lamda = 0.1
        A, x_numpy, y = self.Ax_y_setup(n, lamda)

        def gradf(x):
            gradf_x = A.T @ (A @ x - y)
            gradf_x += lamda * x

            return gradf_x

        def inv_hessf(x):
            I = np.eye(n)
            return lambda x: np.linalg.pinv(A.T @ A + lamda * I) @ x
        for beta in [1, 0.5]:
            with self.subTest(beta=beta):
                if beta < 1:
                    def f(x):
                        f_x = 1 / 2 * np.linalg.norm(A @ x - y)**2
                        f_x += lamda / 2 * np.linalg.norm(x)**2

                        return f_x
                else:
                    f = None

                x = np.zeros(n)
                alg_method = alg.NewtonsMethod(
                    gradf, inv_hessf, x,
                    beta=beta, f=f)
                while (not alg_method.done()):
                    alg_method.update()

                npt.assert_allclose(x, x_numpy)



# -*- coding: utf-8 -*-
"""This module provides an abstract class Alg for iterative algorithms,
and implements commonly used methods.
"""
import numpy as np
from sigpy import backend, util


class Alg(object):
    """Abstraction for iterative algorithms.

    The standard way of using an :class:`Alg` object, say alg, is as follows:

    >>> while not alg.done():
    >>>     alg.update()

    The user is free to run other things in the while loop.
    An :class:`Alg` object is meant to run once.
    Once done, the object should not be run again.

    When creating a new :class:`Alg` class, the user should supply
    an _update() function
    to perform the iterative update, and optionally a _done() function
    to determine when to terminate the iteration. The default _done() function
    simply checks whether the number of iterations has reached the maximum.

    The interface for each :class:`Alg` class should not depend on
    Linop or Prox explicitly.
    For example, if the user wants to design an
    :class:`Alg` class to accept a Linop, say A,
    as an argument, then it should also accept any function that can be called
    to compute x -> A(x). Similarly, to accept a Prox, say proxg,
    as an argument,
    the Alg class should accept any function that can be called to compute
    alpha, x -> proxg(x).

    Args:
        max_iter (int): Maximum number of iterations.

    """

    def __init__(self, max_iter):
        self.max_iter = max_iter
        self.iter = 0

    def _update(self):
        raise NotImplementedError

    def _done(self):
        return self.iter >= self.max_iter

    def update(self):
        self._update()
        self.iter += 1

    def done(self):
        return self._done()


class PowerMethod(Alg):
    """Power method to estimate maximum eigenvalue and eigenvector.

    Args:
        A (Linop or function): Function to a hermitian linear mapping.
        x (array): Variable to optimize over.
        max_iter (int): Maximum number of iterations.

    Attributes:
        max_eig (float): Maximum eigenvalue of `A`.

    """

    def __init__(self, A, x, max_iter=30):
        self.A = A
        self.x = x
        self.max_eig = np.infty
        super().__init__(max_iter)

    def _update(self):
        y = self.A(self.x)
        device = backend.get_device(y)
        xp = device.xp
        with device:
            self.max_eig = util.asscalar(xp.linalg.norm(y))
            backend.copyto(self.x, y / self.max_eig)

    def _done(self):
        return self.iter >= self.max_iter or self.max_eig == 0


class GradientMethod(Alg):
    r"""First order gradient method.

    For the simplest setting when proxg is not specified,
    the method considers the objective function:

    .. math:: \min_x f(x)

    where :math:`f` is (sub)-differentiable and performs the update:

    .. math:: x_\text{new} = x - \alpha \nabla f(x)

    When proxg is specified, the method considers the composite
    objective function:

    .. math:: f(x) + g(x)

    where :math:`f` is (sub)-differentiable and :math:`g` is simple,
    and performs the update:

    .. math:: x_\text{new} = \text{prox}_{\alpha g}(x - \alpha \nabla f(x))

    Nesterov's acceleration is supported by toggling the `accelerate`
    input option.

    Backtracking line search is supported by setting :math:`\beta < 1`,
    which keeps scaling the step-size :math:`\alpha` by :math:`\beta`
    until the following condition holds:

    .. math:: f(x_\text{new}) \leq f(x) +
        \left< \Delta x, \nabla f(x) \right> +
        \frac{1}{2 \alpha} \| \Delta x \|_2^2

    Args:
        gradf (function): function to compute :math:`\nabla f`.
        x (array): variable to optimize over.
        alpha (float or None): step size, or initial step size
             if backtracking line-search is on.
        beta (scalar): backtracking linesearch factor.
             Enables backtracking when beta < 1.
        f (function or None): function to compute :math:`f`
             for backtracking line-search.
        proxg (Prox, function or None): Prox or function to compute
            proximal operator of :math:`g`.
        accelerate (bool): toggle Nesterov acceleration.
        max_iter (int): maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    References:
        Nesterov, Y. E. (1983).
        A method for solving the convex programming problem
        with convergence rate O (1/k^ 2).
        In Dokl. Akad. Nauk SSSR (Vol. 269, pp. 543-547).

        Beck, A., & Teboulle, M. (2009).
        A fast iterative shrinkage-thresholding algorithm
        for linear inverse problems.
        SIAM journal on imaging sciences, 2(1), 183-202.

    """

    def __init__(self, gradf, x, alpha, proxg=None,
                 f=None, beta=1, accelerate=False, max_iter=100,
                 tol=0):
        if beta < 1 and f is None:
            raise TypeError(
                "Cannot do backtracking linesearch without specifying f.")

        self.gradf = gradf
        self.alpha = alpha
        self.f = f
        self.beta = beta
        self.accelerate = accelerate
        self.proxg = proxg
        self.x = x
        self.tol = tol

        self.device = backend.get_device(x)
        with self.device:
            if self.accelerate:
                self.z = self.x.copy()
                self.t = 1

        self.resid = np.infty
        super().__init__(max_iter)

    def _update(self):
        xp = self.device.xp
        with self.device:
            if self.accelerate:
                backend.copyto(self.x, self.z)

            # Perform update
            gradf_x = self.gradf(self.x)
            alpha = self.alpha
            x_new = self.x - alpha * gradf_x
            if self.proxg is not None:
                x_new = self.proxg(alpha, x_new)

            delta_x = x_new - self.x
            # Backtracking line search
            if self.beta < 1:
                fx = self.f(self.x)
                while self.f(x_new) > fx + util.asscalar(
                        xp.real(xp.vdot(delta_x, gradf_x))) + \
                        1 / (2 * alpha) * util.asscalar(
                            xp.linalg.norm(delta_x))**2:
                    alpha *= self.beta

                    x_new = self.x - alpha * gradf_x
                    if self.proxg is not None:
                        x_new = self.proxg(alpha, x_new)

                    delta_x = x_new - self.x

            backend.copyto(self.x, x_new)
            if self.accelerate:
                t_old = self.t
                self.t = (1 + (1 + 4 * t_old**2)**0.5) / 2
                backend.copyto(self.z, x_new +
                               ((t_old - 1) / self.t) * delta_x)

            self.resid = util.asscalar(xp.linalg.norm(delta_x / alpha))

    def _done(self):
        return (self.iter >= self.max_iter) or self.resid <= self.tol


class ConjugateGradient(Alg):
    r"""Conjugate Gradient Method. Solves for:

    .. math:: A x = b

    where A is hermitian.

    Args:
        A (Linop or function): Linop or function to compute A.
        b (array): Observation.
        x (array): Variable.
        P (function or None): Preconditioner.
        max_iter (int): Maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    """

    def __init__(self, A, b, x, P=None, max_iter=100, tol=0):
        self.A = A
        self.P = P
        self.x = x
        self.tol = tol
        self.device = backend.get_device(x)
        with self.device:
            xp = self.device.xp
            self.r = b - self.A(self.x)

            if self.P is None:
                z = self.r
            else:
                z = self.P(self.r)

            if max_iter > 1:
                self.p = z.copy()
            else:
                self.p = z

            self.not_positive_definite = False
            self.rzold = xp.real(xp.vdot(self.r, z))
            self.resid = util.asscalar(self.rzold)**0.5

        super().__init__(max_iter)

    def _update(self):
        with self.device:
            xp = self.device.xp
            Ap = self.A(self.p)
            pAp = util.asscalar(xp.real(xp.vdot(self.p, Ap)))
            if pAp <= 0:
                self.not_positive_definite = True
                return

            self.alpha = self.rzold / pAp
            util.axpy(self.x, self.alpha, self.p)
            if self.iter < self.max_iter - 1:
                util.axpy(self.r, -self.alpha, Ap)
                if self.P is not None:
                    z = self.P(self.r)
                else:
                    z = self.r

                rznew = xp.real(xp.vdot(self.r, z))
                beta = rznew / self.rzold
                util.xpay(self.p, beta, z)
                self.rzold = rznew

            self.resid = util.asscalar(self.rzold)**0.5

    def _done(self):
        return (self.iter >= self.max_iter or
                self.not_positive_definite or self.resid <= self.tol)


class PrimalDualHybridGradient(Alg):
    r"""Primal dual hybrid gradient.

    Considers the problem:

    .. math:: \min_x \max_u - f^*(u) + g(x) + h(x) + <Ax, u>

    Or equivalently:

    .. math:: \min_x f(A x) + g(x) + h(x)

    where f, and g are simple, and h is Lipschitz continuous.

    Args:
        proxfc (function): Function to compute proximal operator of f^*.
        proxg (function): Function to compute proximal operator of g.
        A (function): Function to compute a linear mapping.
        AH (function): Function to compute the adjoint linear mapping of `A`.
        x (array): Primal solution.
        u (array): Dual solution.
        tau (float or array): Primal step-size.
        sigma (float or array): Dual step-size.
        gamma_primal (float): Strong convexity parameter of g.
        gamma_dual (float): Strong convexity parameter of f^*.
        max_iter (int): Maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    References:
       Chambolle, A., & Pock, T. (2011).
       A first-order primal-dual algorithm for convex problems with
       applications to imaging.
       Journal of mathematical imaging and vision, 40(1), 120-145.

    """

    def __init__(self, proxfc, proxg, A, AH, x, u,
                 tau, sigma, theta=1, gradh=None,
                 gamma_primal=0, gamma_dual=0,
                 max_iter=100, tol=0):
        self.proxfc = proxfc
        self.proxg = proxg
        self.gradh = gradh
        self.tol = tol

        self.A = A
        self.AH = AH

        self.u = u
        self.x = x

        self.tau = tau
        self.sigma = sigma
        self.theta = theta
        self.gamma_primal = gamma_primal
        self.gamma_dual = gamma_dual

        self.x_device = backend.get_device(x)
        self.u_device = backend.get_device(u)

        with self.x_device:
            self.x_ext = self.x.copy()

        with self.u_device:
            self.u_old = self.u.copy()
            self.x_old = self.x.copy()

        self.resid = np.infty

        super().__init__(max_iter)

    def _update(self):
        backend.copyto(self.u_old, self.u)
        backend.copyto(self.x_old, self.x)

        # Update dual.
        delta_u = self.A(self.x_ext)
        util.axpy(self.u, self.sigma, delta_u)
        backend.copyto(self.u, self.proxfc(self.sigma, self.u))

        # Update primal.
        with self.x_device:
            delta_x = self.AH(self.u)
            if self.gradh is not None:
                delta_x += self.gradh(self.x)

            util.axpy(self.x, -self.tau, delta_x)
            backend.copyto(self.x, self.proxg(self.tau, self.x))

        # Update step-size if neccessary.
        if self.gamma_primal > 0 and self.gamma_dual == 0:
            with self.x_device:
                xp = self.x_device.xp
                theta = 1 / (1 + 2 * self.gamma_primal *
                             xp.amin(xp.abs(self.tau)))**0.5
                self.tau *= theta

            with self.u_device:
                self.sigma /= theta
        elif self.gamma_primal == 0 and self.gamma_dual > 0:
            with self.u_device:
                xp = self.u_device.xp
                theta = 1 / (1 + 2 * self.gamma_dual *
                             xp.amin(xp.abs(self.sigma)))**0.5
                self.sigma *= theta

            with self.x_device:
                self.tau /= theta
        else:
            theta = self.theta

        # Extrapolate primal.
        with self.x_device:
            xp = self.x_device.xp
            x_diff = self.x - self.x_old
            backend.copyto(self.x_ext, self.x + theta * x_diff)
            x_diff_norm = xp.linalg.norm(x_diff / self.tau**0.5)

        with self.u_device:
            xp = self.u_device.xp
            u_diff = self.u - self.u_old
            u_diff_norm = xp.linalg.norm(u_diff / self.sigma**0.5)

        self.resid = util.asscalar(x_diff_norm**2 + u_diff_norm**2)

    def _done(self):
        return self.iter >= self.max_iter or self.resid <= self.tol


class AltMin(Alg):
    """Alternating Minimization.

    Args:
        min1 (function): Function to minimize over variable 1.
        min2 (function): Funciton to minimize over variable 2.
        max_iter (int): Maximum number of iterations.

    """

    def __init__(self, min1, min2, max_iter=30):
        self.min1 = min1
        self.min2 = min2
        super().__init__(max_iter)

    def _update(self):
        self.min1()
        self.min2()


class AugmentedLagrangianMethod(Alg):
    r"""Augmented Lagrangian method for constrained optimization.

    Consider the problem:

    .. math:: \min_{x: c(x) = 0} f(x)

    and reformulate it as:

    .. math::
        \min_{x, u} L(x, u, \mu) = f(x) - <c(x), u> + \mu / 2 \| c(x) \|_2^2

    Perform the following update steps:

    .. math::
        x \in \text{argmin}_{x} L(x, u, \mu)\\
        u = u - \mu c(x)

    Args:
        min_lagrangian (function): a function that takes :math:`x`, :math:`u`,
            and :math:`\mu` as inputs, and minimizes the Lagrangian over `x`.
        constraints (function): a function that takes :math:`x` as input,
            and outputs :math:`c(x)`.
        x (array): primal variable.
        u (array): dual variable.
        mu (scalar): step size.
        max_iter (int): maximum number of iterations.

    """
    def __init__(self, min_lagrangian, constraints, x, u, mu,
                 max_iter=30):
        self.min_lagrangian = min_lagrangian
        self.constraints = constraints
        self.x = x
        self.u = u
        self.mu = mu
        super().__init__(max_iter)

    def _update(self):
        self.min_lagrangian(self.x, self.u, self.mu)
        util.axpy(self.u, -self.mu, self.constraints(self.x))


class NewtonsMethod(Alg):
    """Newton's Method.

    Args:
        gradf (function) - A function gradf(x): x -> gradient of f at x.
        inv_hessf (function) - A function H(x): x -> inverse Hessian of f at x,
            which is another function: y -> inverse Hessian of f at x times y.
        x (function) - solution.
        beta (scalar): backtracking linesearch factor.
             Enables backtracking when beta < 1.
        f (function or None): function to compute :math:`f`
             for backtracking line-search.
        max_iter (int): maximum number of iterations.
        tol (float): Tolerance for stopping condition.

    """
    def __init__(self, gradf, inv_hessf, x,
                 beta=1, f=None, max_iter=10, tol=0):
        if beta < 1 and f is None:
            raise TypeError(
                "Cannot do backtracking linesearch without specifying f.")

        self.gradf = gradf
        self.inv_hessf = inv_hessf
        self.x = x
        self.lamda = np.infty
        self.beta = beta
        self.f = f
        self.residual = np.infty
        self.tol = tol

        super().__init__(max_iter)

    def _update(self):
        device = backend.get_device(self.x)
        xp = device.xp
        with device:
            gradf_x = self.gradf(self.x)
            p = -self.inv_hessf(self.x)(gradf_x)
            self.lamda2 = util.asscalar(-xp.real(xp.vdot(p, gradf_x)))

            if self.lamda2 < 0:
                raise ValueError(
                    'Hessian is not positive semi-definite. '
                    'inv_hessf might not be defined correctly.')

            x_new = self.x + p
            if self.beta < 1:
                fx = self.f(self.x)
                alpha = 1
                while self.f(x_new) > fx - alpha / 2 * self.lamda2:
                    alpha *= self.beta
                    x_new = self.x + alpha * p

            backend.copyto(self.x, x_new)
            self.residual = self.lamda2**0.5

    def _done(self):
        return self.iter >= self.max_iter or self.residual <= self.tol



# -*- coding: utf-8 -*-
"""MRI utilities.
"""
import numpy as np


__all__ = ['get_cov', 'whiten']


def get_cov(noise):
    """Get covariance matrix from noise measurements.

    Args:
        noise (array): Noise measurements of shape [num_coils, ...]

    Returns:
        array: num_coils x num_coils covariance matrix.

    """
    num_coils = noise.shape[0]
    X = noise.reshape([num_coils, -1])
    X -= np.mean(X, axis=-1, keepdims=True)
    cov = np.matmul(X, X.T.conjugate())

    return cov


def whiten(ksp, cov):
    """Whitens k-space measurements.

    Args:
        ksp (array): k-space measurements of shape [num_coils, ...]
        cov (array): num_coils x num_coils covariance matrix.

    Returns:
        array: whitened k-space array.

    """
    num_coils = ksp.shape[0]

    x = ksp.reshape([num_coils, -1])

    L = np.linalg.cholesky(cov)
    x_w = np.linalg.solve(L, x)
    ksp_w = x_w.reshape(ksp.shape)

    return ksp_w



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import backend, config, linop, pytorch


if config.pytorch_enabled:
    import torch

    if __name__ == '__main__':
        unittest.main()

    devices = [backend.cpu_device]
    if config.cupy_enabled:
        devices.append(backend.Device(0))

    class TestPytorch(unittest.TestCase):

        def test_to_pytorch(self):
            for dtype in [np.float32, np.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        xp = device.xp
                        array = xp.array([1, 2, 3], dtype=dtype)
                        tensor = pytorch.to_pytorch(array)
                        tensor[0] = 0
                        xp.testing.assert_allclose(array, [0, 2, 3])

        def test_to_pytorch_complex(self):
            for dtype in [np.complex64, np.complex128]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        xp = device.xp
                        array = xp.array([1 + 1j, 2 + 2j, 3 + 3j], dtype=dtype)
                        tensor = pytorch.to_pytorch(array)
                        tensor[0, 0] = 0
                        xp.testing.assert_allclose(array, [1j, 2 + 2j, 3 + 3j])

        def test_from_pytorch(self):
            for dtype in [torch.float32, torch.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        if device == backend.cpu_device:
                            torch_device = torch.device('cpu')
                        else:
                            torch_device = torch.device('cuda:0')

                        tensor = torch.tensor([1, 2, 3],
                                              dtype=dtype, device=torch_device)
                        array = pytorch.from_pytorch(tensor)
                        array[0] = 0
                        np.testing.assert_allclose(tensor.cpu().numpy(),
                                                   [0, 2, 3])

        def test_from_pytorch_complex(self):
            for dtype in [torch.float32, torch.float64]:
                for device in devices:
                    with self.subTest(device=device, dtype=dtype):
                        if device == backend.cpu_device:
                            torch_device = torch.device('cpu')
                        else:
                            torch_device = torch.device('cuda:0')

                        tensor = torch.tensor([[1, 1], [2, 2], [3, 3]],
                                              dtype=dtype, device=torch_device)
                        array = pytorch.from_pytorch(tensor, iscomplex=True)
                        xp = device.xp
                        xp.testing.assert_array_equal(array,
                                                      [1 + 1j, 2 + 2j, 3 + 3j])
                        array[0] -= 1
                        np.testing.assert_allclose(tensor.cpu().numpy(),
                                                   [[0, 1], [2, 2], [3, 3]])

        def test_to_pytorch_function(self):
            A = linop.Resize([5], [3])
            x = np.array([1, 2, 3], np.float)
            y = np.ones([5])

            with self.subTest('forward'):
                f = pytorch.to_pytorch_function(A).apply
                x_torch = pytorch.to_pytorch(x)
                npt.assert_allclose(f(x_torch).detach().numpy(),
                                    A(x))

            with self.subTest('adjoint'):
                y_torch = pytorch.to_pytorch(y)
                loss = (f(x_torch) - y_torch).pow(2).sum() / 2
                loss.backward()
                npt.assert_allclose(x_torch.grad.detach().numpy(),
                                    A.H(A(x) - y))

        def test_to_pytorch_function_complex(self):
            A = linop.FFT([3])
            x = np.array([1 + 1j, 2 + 2j, 3 + 3j], np.complex)
            y = np.ones([3], np.complex)

            with self.subTest('forward'):
                f = pytorch.to_pytorch_function(
                    A,
                    input_iscomplex=True,
                    output_iscomplex=True).apply
                x_torch = pytorch.to_pytorch(x)
                npt.assert_allclose(f(x_torch).detach().numpy().ravel(),
                                    A(x).view(np.float))

            with self.subTest('adjoint'):
                y_torch = pytorch.to_pytorch(y)
                loss = (f(x_torch) - y_torch).pow(2).sum() / 2
                loss.backward()
                npt.assert_allclose(x_torch.grad.detach().numpy().ravel(),
                                    A.H(A(x) - y).view(np.float))



# -*- coding: utf-8 -*-
"""Configuration.

This module contains flags to turn on and off optional modules.

"""
from importlib import util

cupy_enabled = util.find_spec("cupy") is not None
if cupy_enabled:  # pragma: no cover
    cudnn_enabled = util.find_spec("cupy.cuda.cudnn") is not None
    nccl_enabled = util.find_spec("cupy.cuda.nccl") is not None
else:
    cudnn_enabled = False
    nccl_enabled = False

mpi4py_enabled = util.find_spec("mpi4py") is not None

pytorch_enabled = util.find_spec("torch") is not None



import pickle
import unittest
import numpy as np
import numpy.testing as npt
from sigpy import util, backend

if __name__ == '__main__':
    unittest.main()


class TestUtil(unittest.TestCase):

    def test_device(self):
        device = backend.Device(-1)
        pickle.dumps(device)

    def test_dirac(self):
        output = util.dirac([5])
        truth = [0, 0, 1, 0, 0]
        npt.assert_allclose(output, truth)

        output = util.dirac([4])
        truth = [0, 0, 1, 0]
        npt.assert_allclose(output, truth)

    def test_triang(self):
        npt.assert_allclose(util.triang([3]), [0.5, 1, 0.5])
        npt.assert_allclose(util.triang([4]), [0.25, 0.75, 0.75, 0.25])

    def test_hanning(self):
        npt.assert_allclose(util.hanning([4]), [0, 0.5, 1, 0.5])
        npt.assert_allclose(util.hanning([5]), [0, 0.5, 1, 0.5, 0])

    def test_resize(self):
        # Zero-pad
        x = np.array([1, 2, 3])
        oshape = [5]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [0, 1, 2, 3, 0])

        x = np.array([1, 2, 3])
        oshape = [4]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [0, 1, 2, 3])

        x = np.array([1, 2])
        oshape = [5]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [0, 1, 2, 0, 0])

        x = np.array([1, 2])
        oshape = [4]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [0, 1, 2, 0])

        # Zero-pad non centered
        x = np.array([1, 2, 3])
        oshape = [5]
        y = util.resize(x, oshape, oshift=[0])
        npt.assert_allclose(y, [1, 2, 3, 0, 0])

        # Crop
        x = np.array([0, 1, 2, 3, 0])
        oshape = [3]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [1, 2, 3])

        x = np.array([0, 1, 2, 3])
        oshape = [3]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [1, 2, 3])

        x = np.array([0, 1, 2, 0, 0])
        oshape = [2]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [1, 2])

        x = np.array([0, 1, 2, 0])
        oshape = [2]
        y = util.resize(x, oshape)
        npt.assert_allclose(y, [1, 2])

        # Crop non centered
        x = np.array([1, 2, 3, 0, 0])
        oshape = [3]
        y = util.resize(x, oshape, ishift=[0])
        npt.assert_allclose(y, [1, 2, 3])

    def test_downsample(self):
        x = np.array([1, 2, 3, 4, 5])
        y = util.downsample(x, [2])
        npt.assert_allclose(y, [1, 3, 5])

    def test_upsample(self):
        x = np.array([1, 2, 3])
        y = util.upsample(x, [5], [2])
        npt.assert_allclose(y, [1, 0, 2, 0, 3])

    def test_circshift(self):
        input = np.array([0, 1, 2, 3])
        axes = [0]
        shift = [1]
        npt.assert_allclose(util.circshift(input, shift, axes),
                            [3, 0, 1, 2])

        input = np.array([[0, 1, 2],
                          [3, 4, 5]])
        axes = [-1]
        shift = [2]
        npt.assert_allclose(util.circshift(input, shift, axes),
                            [[1, 2, 0],
                             [4, 5, 3]])

        input = np.array([[0, 1, 2],
                          [3, 4, 5]])
        axes = [-2]
        shift = [1]
        npt.assert_allclose(util.circshift(input, shift, axes),
                            [[3, 4, 5],
                             [0, 1, 2]])

    def test_monte_carlo_sure(self):
        x = np.ones([100000], dtype=np.float)
        sigma = 0.1
        noise = 0.1 * util.randn([100000], dtype=np.float)
        y = x + noise

        def f(y):
            return y

        npt.assert_allclose(
            sigma**2, util.monte_carlo_sure(f, y, sigma), atol=1e-3)

    def test_ShuffledNumbers(self):
        n = 5
        idx = util.ShuffledNumbers(n)
        x = sorted([idx.next() for _ in range(2 * n)])
        npt.assert_allclose(x, [0, 0, 1, 1, 2, 2, 3, 3, 4, 4])



import unittest
import numpy as np
import sigpy as sp

from sigpy.mri import espirit, sim

if __name__ == '__main__':
    unittest.main()


class TestEspirit(unittest.TestCase):

    def test_espirit_maps(self):
        mps_shape = [8, 32, 32]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2])
        mps_rec = espirit.espirit_maps(ksp)

        np.testing.assert_allclose(np.abs(mps)[:, 8:24, 8:24],
                                   np.abs(mps_rec[:, 8:24, 8:24]),
                                   rtol=1e-3, atol=1e-3)

    def test_espirit_maps_eig(self):
        mps_shape = [8, 32, 32]
        mps = sim.birdcage_maps(mps_shape)
        ksp = sp.fft(mps, axes=[-1, -2])
        mps_rec, eig_val = espirit.espirit_maps(ksp,
                                                output_eigenvalue=True)

        np.testing.assert_allclose(eig_val, 1, rtol=0.01, atol=0.01)



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import prox, util, linop

if __name__ == '__main__':
    unittest.main()


class TestProx(unittest.TestCase):

    def test_L1Reg(self):
        shape = [6]
        lamda = 1.0
        P = prox.L1Reg(shape, lamda)
        phase = np.exp(1j * np.random.random(shape))
        x = np.array([-3.0, -2.0, -1.0, 0, 1.0, 2.0]) * phase
        y = P(1.0, x)
        z = np.array([-2.0, -1.0, -0.0, 0, 0.0, 1.0]) * phase
        npt.assert_allclose(y, z)

    def test_L1Proj(self):
        shape = [6]
        epsilon = 1.0
        P = prox.L1Proj(shape, epsilon)
        x = util.randn(shape)
        y = P(1.0, x)
        z = 1.0 if np.linalg.norm(x, 1) > 1.0 else np.linalg.norm(x, 1)
        npt.assert_allclose(np.linalg.norm(y, 1), z)

        x = util.randn(shape) * 0.0001
        y = P(1.0, x)
        z = 1.0 if np.linalg.norm(x, 1) > 1.0 else np.linalg.norm(x, 1)
        npt.assert_allclose(np.linalg.norm(y, 1), z)

    def test_UnitaryTransform(self):
        shape = [6]
        lamda = 1.0
        A = linop.FFT(shape)
        P = prox.UnitaryTransform(prox.L2Reg(shape, lamda), A)
        x = util.randn(shape)
        y = P(0.1, x)
        npt.assert_allclose(y, x / (1 + lamda * 0.1))

    def test_L2Reg(self):
        shape = [6]
        lamda = 1.0
        P = prox.L2Reg(shape, lamda)
        x = util.randn(shape)
        y = P(0.1, x)
        npt.assert_allclose(y, x / (1 + lamda * 0.1))

    def test_L2Proj(self):
        shape = [6]
        epsilon = 1.0
        P = prox.L2Proj(shape, epsilon)
        x = util.randn(shape)
        y = P(1.0, x)
        npt.assert_allclose(y, x / np.linalg.norm(x.ravel()))

    def test_BoxConstraint(self):
        shape = [5]
        P = prox.BoxConstraint(shape, -1, 1)
        x = np.array([-2, -1, 0, 1, 2])
        y = P(None, x)
        npt.assert_allclose(y, [-1, -1, 0, 1, 1])

        P = prox.BoxConstraint(shape, [-1, 0, -1, -1, -1], [1, 1, 1, 0, 1])
        x = np.array([-2, -1, 0, 1, 2])
        y = P(None, x)
        npt.assert_allclose(y, [-1, 0, 0, 0, 1])



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import app, linop, util, prox

if __name__ == '__main__':
    unittest.main()


class TestApp(unittest.TestCase):

    def test_MaxEig(self):
        n = 5
        mat = util.randn([n, n])
        A = linop.MatMul([n, 1], mat)
        s = np.linalg.svd(mat, compute_uv=False)

        npt.assert_allclose(app.MaxEig(
            A.H * A, max_iter=1000, show_pbar=False).run(), s[0]**2, atol=1e-3)

    def test_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.1 * np.ones([n, n])
        A = linop.MatMul([n, 1], _A)
        x = np.arange(n).reshape([n, 1])
        y = A(x)
        z = np.arange(n).reshape([n, 1])

        for mu in [0, 0.1]:
            for lamda in [0, 0.1]:
                for proxg in [None, prox.L2Reg([n, 1], lamda)]:
                    for alg_name in ['GradientMethod',
                                     'PrimalDualHybridGradient',
                                     'ConjugateGradient']:
                        with self.subTest(proxg=proxg,
                                          alg_name=alg_name,
                                          lamda=lamda,
                                          mu=mu):
                            if proxg is None:
                                prox_lamda = 0
                            else:
                                prox_lamda = lamda

                            x_numpy = np.linalg.solve(
                                _A.T @ _A +
                                (lamda + mu + prox_lamda) * np.eye(n),
                                _A.T @ y + mu * z)

                            if (alg_name == 'ConjugateGradient'
                                and proxg is not None):
                                with self.assertRaises(ValueError):
                                    app.LinearLeastSquares(
                                        A, y,
                                        alg_name=alg_name,
                                        lamda=lamda,
                                        proxg=proxg,
                                        mu=mu, z=z,
                                        show_pbar=False).run()
                            else:
                                x_rec = app.LinearLeastSquares(
                                    A, y,
                                    alg_name=alg_name,
                                    lamda=lamda,
                                    proxg=proxg,
                                    mu=mu, z=z,
                                    show_pbar=False).run()

                                npt.assert_allclose(x_rec, x_numpy, atol=1e-3)

    def test_L2ConstrainedMinimization(self):
        n = 5
        _A = np.eye(n) + 0.1 * util.randn([n, n])
        A = linop.MatMul([n, 1], _A)
        x = util.randn([n, 1])
        y = A(x)

        eps = 0

        def proxg(lamda, x):
            return x / (1 + lamda)

        x_rec = app.L2ConstrainedMinimization(A, y, proxg, eps,
                                              show_pbar=False).run()
        npt.assert_allclose(x_rec, x, atol=1e-3)

    def test_precond_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.1 * util.randn([n, n])
        A = linop.MatMul([n, 1], _A)
        x = util.randn([n, 1])
        y = A(x)
        x_lstsq = np.linalg.lstsq(_A, y, rcond=-1)[0]
        p = 1 / (np.sum(abs(_A)**2, axis=0).reshape([n, 1]))

        P = linop.Multiply([n, 1], p)
        x_rec = app.LinearLeastSquares(A, y, show_pbar=False).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

        alpha = p / app.MaxEig(P * A.H * A, show_pbar=False).run()
        x_rec = app.LinearLeastSquares(
            A,
            y,
            alg_name='GradientMethod',
            alpha=alpha,
            max_power_iter=100,
            max_iter=1000, show_pbar=False).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

        tau = p
        x_rec = app.LinearLeastSquares(
            A,
            y,
            alg_name='PrimalDualHybridGradient',
            max_iter=1000,
            tau=tau, show_pbar=False).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)

    def test_dual_precond_LinearLeastSquares(self):
        n = 5
        _A = np.eye(n) + 0.1 * util.randn([n, n])
        A = linop.MatMul([n, 1], _A)
        x = util.randn([n, 1])
        y = A(x)
        x_lstsq = np.linalg.lstsq(_A, y, rcond=-1)[0]

        d = 1 / np.sum(abs(_A)**2, axis=1, keepdims=True).reshape([n, 1])
        x_rec = app.LinearLeastSquares(
            A,
            y,
            alg_name='PrimalDualHybridGradient',
            max_iter=1000,
            sigma=d, show_pbar=False).run()
        npt.assert_allclose(x_rec, x_lstsq, atol=1e-3)



# -*- coding: utf-8 -*-
"""MRI preconditioners.
"""
import sigpy as sp


__all__ = ['kspace_precond', 'circulant_precond']


def kspace_precond(mps, weights=None, coord=None,
                   lamda=0, device=sp.cpu_device):
    r"""Compute a diagonal preconditioner in k-space.

    Considers the optimization problem:

    .. math::
        \min_P \| P A A^H - I \|_F^2

    where A is the Sense operator.

    Args:
        mps (array): sensitivity maps of shape [num_coils] + image shape.
        weights (array): k-space weights.
        coord (array): k-space coordinates of shape [...] + [ndim].
        lamda (float): regularization.

    Returns:
        array: k-space preconditioner of same shape as k-space.

    """
    dtype = mps.dtype

    if weights is not None:
        weights = sp.to_device(weights, device)

    device = sp.Device(device)
    xp = device.xp

    mps_shape = list(mps.shape)
    img_shape = mps_shape[1:]
    img2_shape = [i * 2 for i in img_shape]
    ndim = len(img_shape)

    scale = sp.prod(img2_shape)**1.5 / sp.prod(img_shape)
    with device:
        if coord is None:
            idx = (slice(None, None, 2), ) * ndim

            ones = xp.zeros(img2_shape, dtype=dtype)
            if weights is None:
                ones[idx] = 1
            else:
                ones[idx] = weights**0.5

            psf = sp.ifft(ones)
        else:
            coord2 = coord * 2
            ones = xp.ones(coord.shape[:-1], dtype=dtype)
            if weights is not None:
                ones *= weights**0.5

            psf = sp.nufft_adjoint(ones, coord2, img2_shape)

        p_inv = []
        for mps_i in mps:
            mps_i = sp.to_device(mps_i, device)
            mps_i_norm2 = xp.linalg.norm(mps_i)**2
            xcorr_fourier = 0
            for mps_j in mps:
                mps_j = sp.to_device(mps_j, device)
                xcorr_fourier += xp.abs(sp.fft(mps_i *
                                               xp.conj(mps_j), img2_shape))**2

            xcorr = sp.ifft(xcorr_fourier)
            xcorr *= psf
            if coord is None:
                p_inv_i = sp.fft(xcorr)[idx]
            else:
                p_inv_i = sp.nufft(xcorr, coord2)

            if weights is not None:
                p_inv_i *= weights**0.5

            p_inv.append(p_inv_i * scale / mps_i_norm2)

        p_inv = (xp.abs(xp.stack(p_inv, axis=0)) + lamda) / (1 + lamda)
        p_inv[p_inv == 0] = 1
        p = 1 / p_inv

        return p.astype(dtype)


def circulant_precond(mps, weights=None, coord=None,
                      lamda=0, device=sp.cpu_device):
    r"""Compute circulant preconditioner.

    Considers the optimization problem:

    .. math::
        \min_P \| A^H A - F P F^H  \|_2^2

    where A is the Sense operator,
    and F is a unitary Fourier transform operator.

    Args:
        mps (array): sensitivity maps of shape [num_coils] + image shape.
        weights (array): k-space weights.
        coord (array): k-space coordinates of shape [...] + [ndim].
        lamda (float): regularization.

    Returns:
        array: circulant preconditioner of image shape.

    """
    if coord is not None:
        coord = sp.to_device(coord, device)

    if weights is not None:
        weights = sp.to_device(weights, device)

    dtype = mps.dtype
    device = sp.Device(device)
    xp = device.xp

    mps_shape = list(mps.shape)
    img_shape = mps_shape[1:]
    img2_shape = [i * 2 for i in img_shape]
    ndim = len(img_shape)

    scale = sp.prod(img2_shape)**1.5 / sp.prod(img_shape)**2
    with device:
        idx = (slice(None, None, 2), ) * ndim
        if coord is None:
            ones = xp.zeros(img2_shape, dtype=dtype)
            if weights is None:
                ones[idx] = 1
            else:
                ones[idx] = weights**0.5

            psf = sp.ifft(ones)
        else:
            coord2 = coord * 2
            ones = xp.ones(coord.shape[:-1], dtype=dtype)
            if weights is not None:
                ones *= weights**0.5

            psf = sp.nufft_adjoint(ones, coord2, img2_shape)

        p_inv = 0
        for mps_i in mps:
            mps_i = sp.to_device(mps_i, device)
            xcorr_fourier = xp.abs(sp.fft(xp.conj(mps_i), img2_shape))**2
            xcorr = sp.ifft(xcorr_fourier)
            xcorr *= psf
            p_inv_i = sp.fft(xcorr)
            p_inv_i = p_inv_i[idx]
            p_inv_i *= scale
            if weights is not None:
                p_inv_i *= weights**0.5

            p_inv += p_inv_i

        p_inv += lamda
        p_inv[p_inv == 0] = 1
        p = 1 / p_inv

        return p.astype(dtype)



# -*- coding: utf-8 -*-
"""Machine learning utilities.
"""
import numpy as np
import sigpy as sp


__all__ = ['labels_to_scores', 'scores_to_labels']


def labels_to_scores(labels):
    """Convert labels to scores.

    Args:
        labels (array): One-dimensional label array.

    Returns:
        array: Score array of shape (len(labels), max(labels) + 1).

    """
    device = sp.get_device(labels)
    xp = device.xp
    with device:
        num_classes = labels.max() + 1
        scores = xp.zeros([len(labels), num_classes], dtype=np.float32)
        scores[xp.arange(len(labels)), labels] = 1

    return scores


def scores_to_labels(scores):
    """Convert scores to labels, by setting peak index to label.

    Args:
        scores (array): Two-dimensional score array.

    Returns:
        array: Label array of lengths scores.shape[0].

    """
    device = sp.get_device(scores)
    xp = device.xp
    with device:
        return xp.argmax(scores, axis=1)



import unittest
import numpy as np
import sigpy as sp
import numpy.testing as npt

from sigpy.mri import linop

if __name__ == '__main__':
    unittest.main()


def check_linop_adjoint(A, dtype=np.float, device=sp.cpu_device):

    device = sp.Device(device)
    x = sp.randn(A.ishape, dtype=dtype, device=device)
    y = sp.randn(A.oshape, dtype=dtype, device=device)

    xp = device.xp
    with device:
        lhs = xp.vdot(A * x, y)
        rhs = xp.vdot(x, A.H * y)

        xp.testing.assert_allclose(lhs, rhs, atol=1e-5, rtol=1e-5)


class TestLinop(unittest.TestCase):

    def test_sense_model(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=np.complex)
        mps = sp.randn(mps_shape, dtype=np.complex)

        mask = np.zeros(img_shape)
        mask[::2, ::2] = 1.0

        A = linop.Sense(mps)

        check_linop_adjoint(A, dtype=np.complex)

        npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]),
                            A * img)

    def test_sense_model_batch(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=np.complex)
        mps = sp.randn(mps_shape, dtype=np.complex)

        mask = np.zeros(img_shape, dtype=np.complex)
        mask[::2, ::2] = 1.0

        A = linop.Sense(mps, coil_batch_size=1)
        check_linop_adjoint(A, dtype=np.complex)
        npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]),
                            A * img)

    def test_noncart_sense_model(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=np.complex)
        mps = sp.randn(mps_shape, dtype=np.complex)

        y, x = np.mgrid[:16, :16]
        coord = np.stack([np.ravel(y - 8), np.ravel(x - 8)], axis=1)

        A = linop.Sense(mps, coord=coord)
        check_linop_adjoint(A, dtype=np.complex)
        npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]).ravel(),
                            (A * img).ravel(), atol=0.1, rtol=0.1)

    def test_noncart_sense_model_batch(self):
        img_shape = [16, 16]
        mps_shape = [8, 16, 16]

        img = sp.randn(img_shape, dtype=np.complex)
        mps = sp.randn(mps_shape, dtype=np.complex)

        y, x = np.mgrid[:16, :16]
        coord = np.stack([np.ravel(y - 8), np.ravel(x - 8)], axis=1)

        A = linop.Sense(mps, coord=coord, coil_batch_size=1)
        check_linop_adjoint(A, dtype=np.complex)
        npt.assert_allclose(sp.fft(img * mps, axes=[-1, -2]).ravel(),
                            (A * img).ravel(), atol=0.1, rtol=0.1)

    if sp.config.mpi4py_enabled:
        def test_sense_model_with_comm(self):
            img_shape = [16, 16]
            mps_shape = [8, 16, 16]
            comm = sp.Communicator()

            img = sp.randn(img_shape, dtype=np.complex)
            mps = sp.randn(mps_shape, dtype=np.complex)
            comm.allreduce(img)
            comm.allreduce(mps)
            ksp = sp.fft(img * mps, axes=[-1, -2])

            A = linop.Sense(mps[comm.rank::comm.size], comm=comm)

            npt.assert_allclose(A.H(ksp[comm.rank::comm.size]), np.sum(
                sp.ifft(ksp, axes=[-1, -2]) * mps.conjugate(), 0))



# -*- coding: utf-8 -*-
"""FFT functions.

This module contains FFT functions that support centered operation.

"""
import numpy as np

from sigpy import backend, interp, util


__all__ = ['fft', 'ifft', 'nufft', 'nufft_adjoint', 'estimate_shape']


def fft(input, oshape=None, axes=None, center=True, norm='ortho'):
    """FFT function that supports centering.

    Args:
        input (array): input array.
        oshape (None or array of ints): output shape.
        axes (None or array of ints): Axes over which to compute the FFT.
        norm (Nonr or ``"ortho"``): Keyword to specify the normalization mode.

    Returns:
        array: FFT result of dimension oshape.

    See Also:
        :func:`numpy.fft.fftn`

    """
    device = backend.get_device(input)
    xp = device.xp

    with device:
        if not np.issubdtype(input.dtype, np.complexfloating):
            input = input.astype(np.complex)

        if center:
            output = _fftc(input, oshape=oshape, axes=axes, norm=norm)
        else:
            output = xp.fft.fftn(input, s=oshape, axes=axes, norm=norm)

        if np.issubdtype(input.dtype,
                         np.complexfloating) and input.dtype != output.dtype:
            output = output.astype(input.dtype, copy=False)

        return output


def ifft(input, oshape=None, axes=None, center=True, norm='ortho'):
    """IFFT function that supports centering.

    Args:
        input (array): input array.
        oshape (None or array of ints): output shape.
        axes (None or array of ints): Axes over which to compute
            the inverse FFT.
        norm (None or ``"ortho"``): Keyword to specify the normalization mode.

    Returns:
        array of dimension oshape.

    See Also:
        :func:`numpy.fft.ifftn`

    """
    device = backend.get_device(input)
    xp = device.xp

    with device:
        if not np.issubdtype(input.dtype, np.complexfloating):
            input = input.astype(np.complex)

        if center:
            output = _ifftc(input, oshape=oshape, axes=axes, norm=norm)
        else:
            output = xp.fft.ifftn(input, s=oshape, axes=axes, norm=norm)

        if np.issubdtype(input.dtype,
                         np.complexfloating) and input.dtype != output.dtype:
            output = output.astype(input.dtype)

        return output


def nufft(input, coord, oversamp=1.25, width=4.0, n=128):
    """Non-uniform Fast Fourier Transform.

    Args:
        input (array): input signal domain array of shape
            (..., n_{ndim - 1}, ..., n_1, n_0),
            where ndim is specified by coord.shape[-1]. The nufft
            is applied on the last ndim axes, and looped over
            the remaining axes.
        coord (array): Fourier domain coordinate array of shape (..., ndim).
            ndim determines the number of dimensions to apply the nufft.
            coord[..., i] should be scaled to have its range between
            -n_i // 2, and n_i // 2.
        oversamp (float): oversampling factor.
        width (float): interpolation kernel full-width in terms of
            oversampled grid.
        n (int): number of sampling points of the interpolation kernel.

    Returns:
        array: Fourier domain data of shape
            input.shape[:-ndim] + coord.shape[:-1].

    References:
        Fessler, J. A., & Sutton, B. P. (2003).
        Nonuniform fast Fourier transforms using min-max interpolation
        IEEE Transactions on Signal Processing, 51(2), 560-574.
        Beatty, P. J., Nishimura, D. G., & Pauly, J. M. (2005).
        Rapid gridding reconstruction with a minimal oversampling ratio.
        IEEE transactions on medical imaging, 24(6), 799-808.

    """
    device = backend.get_device(input)
    ndim = coord.shape[-1]
    beta = np.pi * (((width / oversamp) * (oversamp - 0.5))**2 - 0.8)**0.5
    os_shape = _get_oversamp_shape(input.shape, ndim, oversamp)

    with device:
        output = input.copy()

        # Apodize
        _apodize(output, ndim, oversamp, width, beta)

        # Zero-pad
        output /= util.prod(input.shape[-ndim:])**0.5
        output = util.resize(output, os_shape)

        # FFT
        output = fft(output, axes=range(-ndim, 0), norm=None)

        # Interpolate
        coord = _scale_coord(backend.to_device(
            coord, device), input.shape, oversamp)
        kernel = _get_kaiser_bessel_kernel(n, width, beta, coord.dtype, device)
        output = interp.interpolate(output, width, kernel, coord)

        return output


def estimate_shape(coord):
    """Estimate array shape from coordinates.

    Shape is estimated by the different between maximum and minimum of
    coordinates in each axis.

    Args:
        coord (array): Coordinates.
    """
    ndim = coord.shape[-1]
    with backend.get_device(coord):
        shape = [int(coord[..., i].max() - coord[..., i].min())
                 for i in range(ndim)]

    return shape


def nufft_adjoint(input, coord, oshape=None, oversamp=1.25, width=4.0, n=128):
    """Adjoint non-uniform Fast Fourier Transform.

    Args:
        input (array): input Fourier domain array of shape
            (...) + coord.shape[:-1]. That is, the last dimensions
            of input must match the first dimensions of coord.
            The nufft_adjoint is applied on the last coord.ndim - 1 axes,
            and looped over the remaining axes.
        coord (array): Fourier domain coordinate array of shape (..., ndim).
            ndim determines the number of dimension to apply nufft adjoint.
            coord[..., i] should be scaled to have its range between
            -n_i // 2, and n_i // 2.
        oshape (tuple of ints): output shape of the form
            (..., n_{ndim - 1}, ..., n_1, n_0).
        oversamp (float): oversampling factor.
        width (float): interpolation kernel full-width in terms of
            oversampled grid.
        n (int): number of sampling points of the interpolation kernel.

    Returns:
        array: signal domain array with shape specified by oshape.

    See Also:
        :func:`sigpy.nufft.nufft`

    """
    device = backend.get_device(input)
    ndim = coord.shape[-1]
    beta = np.pi * (((width / oversamp) * (oversamp - 0.5))**2 - 0.8)**0.5
    if oshape is None:
        oshape = list(input.shape[:-coord.ndim + 1]) + estimate_shape(coord)
    else:
        oshape = list(oshape)

    os_shape = _get_oversamp_shape(oshape, ndim, oversamp)

    with device:
        # Gridding
        coord = _scale_coord(backend.to_device(
            coord, device), oshape, oversamp)
        kernel = _get_kaiser_bessel_kernel(n, width, beta, coord.dtype, device)
        output = interp.gridding(input, os_shape, width, kernel, coord)

        # IFFT
        output = ifft(output, axes=range(-ndim, 0), norm=None)

        # Crop
        output = util.resize(output, oshape)
        output *= util.prod(os_shape[-ndim:]) / util.prod(oshape[-ndim:])**0.5

        # Apodize
        _apodize(output, ndim, oversamp, width, beta)

        return output


def _fftc(input, oshape=None, axes=None, norm='ortho'):

    ndim = input.ndim
    axes = util._normalize_axes(axes, ndim)
    device = backend.get_device(input)
    xp = device.xp

    if oshape is None:
        oshape = input.shape

    with device:
        tmp = util.resize(input, oshape)
        tmp = xp.fft.ifftshift(tmp, axes=axes)
        tmp = xp.fft.fftn(tmp, axes=axes, norm=norm)
        output = xp.fft.fftshift(tmp, axes=axes)
        return output


def _ifftc(input, oshape=None, axes=None, norm='ortho'):
    ndim = input.ndim
    axes = util._normalize_axes(axes, ndim)
    device = backend.get_device(input)
    xp = device.xp

    if oshape is None:
        oshape = input.shape

    with device:
        tmp = util.resize(input, oshape)
        tmp = xp.fft.ifftshift(tmp, axes=axes)
        tmp = xp.fft.ifftn(tmp, axes=axes, norm=norm)
        output = xp.fft.fftshift(tmp, axes=axes)
        return output


def _get_kaiser_bessel_kernel(n, width, beta, dtype, device):
    """Precompute Kaiser Bessel kernel.

    Precomputes Kaiser-Bessel kernel with n points.

    Args:
        n (int): number of sampling points.
        width (float): kernel width.
        beta (float): kaiser bessel parameter.
        dtype (dtype): output data type.
        device (Device): output device.

    Returns:
        array: Kaiser-Bessel kernel table.

    """
    device = backend.Device(device)
    xp = device.xp
    with device:
        x = xp.arange(n, dtype=dtype) / n
        kernel = 1 / width * xp.i0(beta * (1 - x**2)**0.5).astype(dtype)
        return kernel


def _scale_coord(coord, shape, oversamp):
    ndim = coord.shape[-1]
    device = backend.get_device(coord)
    scale = backend.to_device(
        [_get_ugly_number(oversamp * i) / i for i in shape[-ndim:]], device)
    shift = backend.to_device(
        [_get_ugly_number(oversamp * i) // 2 for i in shape[-ndim:]], device)

    with device:
        coord = scale * coord + shift

    return coord


def _get_ugly_number(n):
    """Get closest ugly number greater than n.

    An ugly number is defined as a positive integer that is a
    multiple of 2 3 and 5.

    Args:
        n (int): Base number.

    """
    if n <= 1:
        return n

    ugly_nums = [1]
    i2, i3, i5 = 0, 0, 0
    while(True):

        ugly_num = min(ugly_nums[i2] * 2,
                       ugly_nums[i3] * 3,
                       ugly_nums[i5] * 5)

        if ugly_num >= n:
            return ugly_num

        ugly_nums.append(ugly_num)
        if ugly_num == ugly_nums[i2] * 2:
            i2 += 1
        elif ugly_num == ugly_nums[i3] * 3:
            i3 += 1
        elif ugly_num == ugly_nums[i5] * 5:
            i5 += 1


def _get_oversamp_shape(shape, ndim, oversamp):
    return list(shape)[:-ndim] + [_get_ugly_number(oversamp * i)
                                  for i in shape[-ndim:]]


def _apodize(input, ndim, oversamp, width, beta):
    device = backend.get_device(input)
    xp = device.xp

    output = input
    with device:
        for a in range(-ndim, 0):
            i = output.shape[a]
            os_i = _get_ugly_number(oversamp * i)
            idx = xp.arange(i, dtype=output.dtype)

            # Calculate apodization
            apod = (beta**2 - (np.pi * width * (idx - i // 2) / os_i)**2)**0.5
            apod /= xp.sinh(apod)
            output *= apod.reshape([i] + [1] * (-a - 1))

        return output



import sys
from setuptools import setup, find_packages

if sys.version_info < (3, 5):
    sys.exit('Sorry, Python < 3.5 is not supported')

REQUIRED_PACKAGES = ['numpy', 'pywavelets', 'numba', 'tqdm']

with open("README.rst", "r") as f:
    long_description = f.read()

setup(name='sigpy',
      version='0.1.8',
      description='Python package for signal reconstruction.',
      long_description=long_description,
      long_description_content_type="text/x-rst",
      url='http://github.com/mikgroup/sigpy',
      author='Frank Ong',
      author_email='frankong@berkeley.edu',
      license='BSD',
      packages=find_packages(),
      install_requires=REQUIRED_PACKAGES,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: BSD License",
          "Operating System :: OS Independent",
      ]
      )



# -*- coding: utf-8 -*-
"""Machine learning applications.
"""
import pickle
import pathlib
import sigpy as sp


class ConvSparseDecom(sp.app.LinearLeastSquares):
    r"""Convolutional sparse decomposition app.

    Considers the convolutional sparse linear model
    :math:`y = \sum_j L_j * R_j`,
    with :math:`L` fixed, and the problem,

    .. math::
        \min_{R} \frac{1}{2}\|y - \sum_j L_j * R_j\|_2^2 + \lambda \|R\|_1

    Args:
        y (array): data array, the first dimension is the number of data.
            If multi_channel is True,
            then the second dimension should be the number of channels.
        L (array): filter array. If multi_channel is True,
            the first dimension is the number of filters.
            Otherwise, the first dimension
            is the number of channels, and second dimension
            is the number of filters.
        lamda (float): regularization parameter.
        mode (str): convolution mode in forward model. {'full', 'valid'}.
            If 'full', then R is smaller than y. If 'valid',
            then R is larger than y.
        multi_channel (bool): whether data is multi-channel or not.
        **kwargs: other LinearLeastSquares arguments.

    Returns:
        array: Coefficients. First two dimensions are number of data
        and number of filters.

    See Also:
        :func:`sigpy.app.LinearLeastSquares`

    """

    def __init__(
            self,
            y,
            L,
            lamda=0.005,
            mode='full',
            multi_channel=False,
            device=sp.cpu_device,
            **kwargs):
        self.y = sp.to_device(y, device)
        self.L = sp.to_device(L, device)
        self.lamda = lamda
        self.mode = mode
        self.multi_channel = multi_channel
        self.device = device

        self._get_params()
        self.A_R = sp.linop.ConvolveInput(
            self.R_shape,
            self.L,
            mode=self.mode,
            input_multi_channel=True,
            output_multi_channel=self.multi_channel)

        proxg_R = sp.prox.L1Reg(self.R_shape, lamda)
        super().__init__(self.A_R, self.y, proxg=proxg_R, **kwargs)

    def _get_params(self):
        self.device = sp.Device(self.device)
        self.dtype = self.y.dtype
        self.num_data = len(self.y)
        self.filt_width = self.L.shape[-1]
        self.num_filters = self.L.shape[self.multi_channel]
        self.data_ndim = self.y.ndim - self.multi_channel - 1

        if self.mode == 'full':
            self.R_shape = (
                [self.num_data, self.num_filters] +
                [i - self.filt_width + 1
                 for i in self.y.shape[-self.data_ndim:]])
        else:
            self.R_shape = (
                [self.num_data, self.num_filters] +
                [i + self.filt_width - 1
                 for i in self.y.shape[-self.data_ndim:]])


class ConvSparseCoding(sp.app.App):
    r"""Convolutional sparse coding application.

    Considers the convolutional sparse model
    :math:`y_t = \sum_j L_j * R_{tj}`,
    and the objective function

    .. math::
        f(L, R) = \sum_t \frac{1}{2} \|y_t - \sum_j L_j * R_{tj}\|_2^2 +
        \lambda \|R\|_1

    where :math:`y_t` is the tth data,
    :math:`L_j` is the jth filter constrained to have unit norm,
    and :math:`R_{tj}` is the jth coefficient for t th data.

    Args:
        y (array): data array, the first dimension is the number of data.
            If multi_channel is True,
            then the second dimension should be the number of channels.
        num_filters (int): number of filters.
        filt_width (int): filter width.
        batch_size (int): batch size.
        lamda (float): regularization parameter.
        alpha (float): step-size.
        max_inner_iter (int): maximum number of iteration for inner-loop.
        max_power_iter (int): maximum number of iteration for power method.
        mode (str): convolution mode in forward model. {'full', 'valid'}.
        multi_channel (bool): whether data is multi-channel or not.
        **kwargs: other LinearLeastSquares arguments.

    Returns:
        array: Filters.

    See Also:
        :func:`sigpy.learn.app.ConvSparseDecom`

    References:
        TODO

    """

    def __init__(self, y, num_filters, filt_width, batch_size,
                 lamda=0.001, alpha=0.5,
                 max_inner_iter=100,
                 max_power_iter=10,
                 max_iter=10,
                 mode='full',
                 multi_channel=False,
                 device=sp.cpu_device,
                 checkpoint_path=None, show_pbar=True):
        self.y = y
        self.num_filters = num_filters
        self.filt_width = filt_width
        self.batch_size = batch_size
        self.lamda = lamda
        self.alpha = alpha
        self.max_inner_iter = max_inner_iter
        self.max_power_iter = max_power_iter
        self.max_iter = max_iter
        self.mode = mode
        self.multi_channel = multi_channel
        self.device = device
        self.checkpoint_path = checkpoint_path

        self._get_params()
        self._get_vars()
        self._get_alg()

        super().__init__(self.alg, show_pbar=show_pbar)

    def _get_params(self):
        self.device = sp.Device(self.device)
        self.dtype = self.y.dtype
        self.data_ndim = self.y.ndim - self.multi_channel - 1
        if self.checkpoint_path is not None:
            self.checkpoint_path = pathlib.Path(self.checkpoint_path)
            self.checkpoint_path.mkdir(parents=True, exist_ok=True)

        self.batch_size = min(len(self.y), self.batch_size)
        self.num_batches = len(self.y) // self.batch_size

        self.L_shape = [self.num_filters] + [self.filt_width] * self.data_ndim
        if self.multi_channel:
            self.L_shape = [self.y.shape[1]] + self.L_shape

        if self.mode == 'full':
            self.R_t_shape = (
                [self.batch_size, self.num_filters] +
                [i - self.filt_width + 1
                 for i in self.y.shape[-self.data_ndim:]])
        else:
            self.R_t_shape = (
                [self.batch_size, self.num_filters] +
                [i + self.filt_width - 1
                 for i in self.y.shape[-self.data_ndim:]])

    def _get_vars(self):
        self.t_idx = sp.ShuffledNumbers(self.num_batches)
        xp = self.device.xp
        with self.device:
            self.y_t = xp.empty((self.batch_size, ) +
                                self.y.shape[1:], dtype=self.dtype)
            self.L = sp.randn(self.L_shape, dtype=self.dtype,
                              device=self.device)
            if self.multi_channel:
                self.L /= xp.sum(xp.abs(self.L)**2, axis=(0, ) +
                                 tuple(range(-self.data_ndim, 0)),
                                 keepdims=True)**0.5
            else:
                self.L /= xp.sum(xp.abs(self.L)**2,
                                 axis=tuple(range(-self.data_ndim,
                                                  0)),
                                 keepdims=True)**0.5

            self.L_old = xp.empty(self.L_shape, dtype=self.dtype)
            self.R = ConvSparseCoefficients(
                self.y,
                self.L,
                lamda=self.lamda,
                multi_channel=self.multi_channel,
                mode=self.mode,
                max_iter=self.max_inner_iter,
                max_power_iter=self.max_power_iter)

    def _get_alg(self):
        def min_R_t():
            self.R_t = ConvSparseDecom(
                self.y_t,
                self.L,
                lamda=self.lamda,
                mode=self.mode,
                multi_channel=self.multi_channel,
                max_power_iter=self.max_power_iter,
                max_iter=self.max_inner_iter,
                device=self.device).run()

        def min_L():
            self.A_L = sp.linop.ConvolveFilter(
                self.L_shape,
                self.R_t,
                mode=self.mode,
                input_multi_channel=True,
                output_multi_channel=self.multi_channel)

            mu = (1 - self.alpha) / self.alpha
            if self.multi_channel:
                proxg_L = sp.prox.L2Proj(self.L.shape, 1, axes=[
                                         0] + list(range(-self.data_ndim, 0)))
            else:
                proxg_L = sp.prox.L2Proj(
                    self.L.shape, 1, axes=range(-self.data_ndim, 0))

            sp.app.LinearLeastSquares(self.A_L, self.y_t, x=self.L,
                                      mu=mu, z=self.L_old,
                                      proxg=proxg_L,
                                      max_power_iter=self.max_power_iter,
                                      max_iter=self.max_inner_iter).run()

        self.alg = sp.alg.AltMin(min_R_t, min_L, max_iter=self.max_iter)

    def _pre_update(self):
        t = self.t_idx.next()
        t_start = t * self.batch_size
        t_end = (t + 1) * self.batch_size

        sp.copyto(self.y_t, self.y[t_start:t_end])
        sp.copyto(self.L_old, self.L)

    def _summarize(self):
        if self.checkpoint_path is not None:
            xp = self.device.xp
            with self.device:
                xp.save(str(self.checkpoint_path / 'L.npy'), self.L)
                xp.save(str(self.checkpoint_path / 'R_t.npy'), self.R_t)
                x = sp.convolve(
                    self.R_t,
                    self.L,
                    mode=self.mode,
                    input_multi_channel=True,
                    output_multi_channel=self.multi_channel)
                xp.save(str(self.checkpoint_path / 'x_t.npy'), x)
                xp.save(str(self.checkpoint_path / 'y_t.npy'), self.y_t)

    def _output(self):
        R = ConvSparseCoefficients(
            self.y,
            self.L,
            lamda=self.lamda,
            multi_channel=self.multi_channel,
            mode=self.mode,
            max_iter=self.max_inner_iter,
            max_power_iter=self.max_power_iter)
        return self.L, R


class LinearRegression(sp.app.App):
    r"""Performs linear regression to fit input to output.

    Considers the linear model :math:`y_t = M x_t`, and the problem,

    .. math::
        \min_M \sum_t \frac{1}{2} \| y_t - M x_t \|_2^2

    where :math:`y_t` is the tth output, :math:`M` is the learned matrix,
    and :math:`x_t` is the tth input.

    It uses the randomized block Kaczmarz method.

    Args:
        input (array): input data of shape (num_data, ...).
        output (array): output data of shape (num_data, ...).
        batch_size (int): batch size.
        mu (float): step size.

    Returns:
       array: matrix of shape input.shape[1:] + output.shape[1:].

    References:
        Needell, Deanna, Ran Zhao, and Anastasios Zouzias.
        Randomized block Kaczmarz method with projection
        for solving least squares.
        Linear Algebra and its Applications 484 (2015): 322-343.

    """

    def __init__(self, input, output, batch_size, mu,
                 lamda=0,
                 max_iter=100, max_inner_iter=100, device=sp.cpu_device,
                 checkpoint_path=None):
        dtype = output.dtype

        num_data = len(output)
        num_batches = num_data // batch_size
        self.device = device
        self.lamda = lamda
        self.batch_size = batch_size
        self.input = input
        self.output = output
        self.checkpoint_path = checkpoint_path

        self.device = sp.Device(device)
        xp = self.device.xp
        with self.device:
            self.mat = xp.zeros(
                input.shape[1:] + output.shape[1:], dtype=dtype)
            self.input_t = xp.empty(
                (batch_size, ) + input.shape[1:], dtype=dtype)
            self.output_t = xp.empty(
                (batch_size, ) + output.shape[1:], dtype=dtype)
            self.t_idx = sp.ShuffledNumbers(num_batches)

        self._get_A()

        def proxf(mu, x):
            return sp.app.LinearLeastSquares(self.A, self.output_t, x=x,
                                             lamda=self.lamda / num_batches,
                                             mu=1 / mu, z=x,
                                             max_iter=max_inner_iter).run()

        alg = sp.alg.ProximalPointMethod(
            proxf, mu, self.mat, max_iter=max_iter)
        super().__init__(alg)

    def _pre_update(self):
        t = self.t_idx.next()
        t_start = t * self.batch_size
        t_end = (t + 1) * self.batch_size

        sp.copyto(self.input_t, self.input[t_start:t_end])
        sp.copyto(self.output_t, self.output[t_start:t_end])

    def _summarize(self):
        xp = self.device.xp
        if self.checkpoint_path is not None:
            xp.save(self.checkpoint_path, self.mat)

    def _output(self):
        return self.mat

    def _get_A(self):
        input_t_size = sp.prod(self.input_t.shape[1:])
        output_t_size = sp.prod(self.output_t.shape[1:])

        Ri = sp.linop.Reshape([input_t_size, output_t_size], self.mat.shape)
        M = sp.linop.MatMul([input_t_size, output_t_size],
                            self.input_t.reshape([self.batch_size, -1]))
        Ro = sp.linop.Reshape(self.output_t.shape, [
                              self.batch_size, output_t_size])
        self.A = Ro * M * Ri


class ConvSparseCoefficients(object):
    r"""Convolutional sparse coefficients.

    Generates coefficients on the fly using convolutional
    sparse decomposition.
    ConvSparseCoefficients can be sliced like arrays.

    Args:
        data (array): data array, the first dimension is the number of data.
            If multi_channel is True, then the second dimension
            should be the number of channels.
        L (array): filter. If multi_channel is True,
            the first dimension is the number of filters.
            Otherwise, the first dimension
            is the number of channels,
            and second dimension is the number of filters.
        lamda (float): regularization parameter.
        mode (str): convolution mode in forward model. {'full', 'valid'}.
        multi_channel (bool): whether data is multi-channel or not.
        max_iter (bool): maximum number of iterations.
        max_power_iter (bool): maximum number of power iterations.

    Attributes:
        shape (tuple of ints): coefficient shape.
        ndim (int): number of dimensions of coefficient.
        dtype (Dtype): Data type.

    """

    def __init__(self, y, L,
                 lamda=1, multi_channel=False, mode='full',
                 max_iter=100, max_power_iter=10,
                 device=sp.cpu_device):

        self.y = y
        self.L = L
        self.lamda = lamda
        self.multi_channel = multi_channel
        self.mode = mode
        self.max_iter = max_iter
        self.max_power_iter = max_power_iter

        self._get_params()
        self.ndim = len(self.shape)
        self.dtype = y.dtype
        self.use_device(device)

    def use_device(self, device):
        self.device = sp.Device(device)

    def __getitem__(self, index):
        if isinstance(index, int):
            y_t = self.y[index:(index + 1)]
            index = 0
        elif isinstance(index, slice):
            y_t = self.y[index]
            index = slice(None)
        elif isinstance(index, tuple) or isinstance(index, list):
            if isinstance(index[0], int):
                y_t = self.y[index[0]:(index[0] + 1)]
                index = [0] + list(index[1:])
            else:
                y_t = self.y[index[0]]
                index = [slice(None)] + list(index[1:])

        with self.device:
            return ConvSparseDecom(
                y_t,
                self.L,
                lamda=self.lamda,
                multi_channel=self.multi_channel,
                mode=self.mode,
                max_iter=self.max_iter,
                max_power_iter=self.max_power_iter,
                device=self.device).run()[index]

    def __len__(self):
        return self.num_data

    def _get_params(self):
        self.num_data = len(self.y)
        if self.multi_channel:
            self.num_filters = self.L.shape[1]
            self.data_ndim = self.y.ndim - 2
        else:
            self.num_filters = self.L.shape[0]
            self.data_ndim = self.y.ndim - 1

        if self.mode == 'full':
            self.shape = tuple(
                [self.num_data, self.num_filters] + [
                    i - min(i, f) + 1 for i, f in zip(
                        self.y.shape[-self.data_ndim:],
                        self.L.shape[-self.data_ndim:])])
        else:
            self.shape = tuple(
                [self.num_data, self.num_filters] + [
                    i + min(i, f) - 1 for i, f in zip(
                        self.y.shape[-self.data_ndim:],
                        self.L.shape[-self.data_ndim:])])

    def save(self, filename):
        with open(filename, "wb") as f:
            pickle.dump(self, f)



"""The core module contains functions and classes for signal processing.

SigPy provides simple interfaces to commonly used signal processing functions,
including convolution, FFT, NUFFT, wavelet transform, and thresholdings.
All functions, except wavelet transform, can run on both CPU and GPU.

These functions are wrapped in two higher level classes to better interface
with iterative methods:
Linop abstracting linear operator and Prox abstracting proximal operator.
SigPy provides an abstraction class Alg for iterative algorithms,
and implements commonly used methods, including conjugate gradient,
(accelerated/proximal) gradient method, and primal dual hybrid gradient.

These classes can then be used to build an App as a final deliverable.
An App simplifies the usage of Alg, and provides convenient features
such as progress bars. A particularly useful App implemented
is the LinearLeastSquares App.

"""
from sigpy import alg, app, config, linop, prox

from sigpy import (backend, block, conv, interp,
                   fourier, pytorch, sim, thresh, util)
from sigpy.backend import *  # noqa
from sigpy.block import *  # noqa
from sigpy.conv import *  # noqa
from sigpy.interp import *  # noqa
from sigpy.fourier import *  # noqa
from sigpy.pytorch import * # noqa
from sigpy.sim import *  # noqa
from sigpy.thresh import *  # noqa
from sigpy.util import *  # noqa

__all__ = ['alg', 'app', 'config', 'linop', 'prox']
__all__.extend(backend.__all__)
__all__.extend(block.__all__)
__all__.extend(conv.__all__)
__all__.extend(interp.__all__)
__all__.extend(fourier.__all__)
__all__.extend(pytorch.__all__)
__all__.extend(sim.__all__)
__all__.extend(thresh.__all__)
__all__.extend(util.__all__)



import unittest
import numpy as np
import numpy.testing as npt
from sigpy import backend, conv, util, config


if __name__ == '__main__':
    unittest.main()


class TestConv(unittest.TestCase):

    def test_convolve_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([1, 3], dtype=dtype)
                    y = backend.to_device(conv.convolve(
                        x, W, mode=mode), backend.cpu_device)
                    npt.assert_allclose(y, [[1]], atol=1e-5)

                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([1, 2], dtype=dtype)
                    y = backend.to_device(conv.convolve(
                        x, W, mode=mode), backend.cpu_device)
                    npt.assert_allclose(y, [[1, 1]], atol=1e-5)

                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([2, 1, 3], dtype=dtype)
                    y = backend.to_device(
                        conv.convolve(
                            x,
                            W,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(y, [[[1]],
                                            [[1]]], atol=1e-5)

    def test_convolve_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([1, 3], dtype=dtype)
                    y = backend.to_device(conv.convolve(
                        x, W, mode=mode), backend.cpu_device)
                    npt.assert_allclose(y, [[0, 1, 1, 1, 0]], atol=1e-5)

                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([1, 2], dtype=dtype)
                    y = backend.to_device(conv.convolve(
                        x, W, mode=mode), backend.cpu_device)
                    npt.assert_allclose(y, [[0, 1, 1, 0]], atol=1e-5)

                    x = util.dirac([1, 3], device=device, dtype=dtype)
                    W = xp.ones([2, 1, 3], dtype=dtype)
                    y = backend.to_device(
                        conv.convolve(
                            x,
                            W,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(y, [[[0, 1, 1, 1, 0]],
                                            [[0, 1, 1, 1, 0]]], atol=1e-5)

    def test_convolve_adjoint_input_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    y = xp.ones([1, 1], dtype=dtype)
                    W = xp.ones([1, 3], dtype=dtype)
                    x = backend.to_device(conv.convolve_adjoint_input(
                        W, y, mode=mode), backend.cpu_device)
                    npt.assert_allclose(x, [[1, 1, 1]], atol=1e-5)

                    y = xp.ones([1, 2], dtype=dtype)
                    W = xp.ones([1, 2], dtype=dtype)
                    x = backend.to_device(conv.convolve_adjoint_input(
                        W, y, mode=mode), backend.cpu_device)
                    npt.assert_allclose(x, [[1, 2, 1]], atol=1e-5)

                    y = xp.ones([2, 1, 1], dtype=dtype)
                    W = xp.ones([2, 1, 3], dtype=dtype)
                    x = backend.to_device(
                        conv.convolve_adjoint_input(
                            W,
                            y,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(x, [[2, 2, 2]], atol=1e-5)

    def test_convolve_adjoint_input_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    y = xp.ones([1, 5], dtype=dtype)
                    W = xp.ones([1, 3], dtype=dtype)
                    x = backend.to_device(conv.convolve_adjoint_input(
                        W, y, mode=mode), backend.cpu_device)
                    npt.assert_allclose(x, [[3, 3, 3]], atol=1e-5)

                    y = xp.ones([1, 4], dtype=dtype)
                    W = xp.ones([1, 2], dtype=dtype)
                    x = backend.to_device(
                        conv.convolve_adjoint_input(
                            W, y, mode=mode), backend.cpu_device)
                    npt.assert_allclose(x, [[2, 2, 2]], atol=1e-5)

                    y = xp.ones([2, 1, 5], dtype=dtype)
                    W = xp.ones([2, 1, 3], dtype=dtype)
                    x = backend.to_device(
                        conv.convolve_adjoint_input(
                            W,
                            y,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(x, [[6, 6, 6]], atol=1e-5)

    def test_convolve_adjoint_filter_valid(self):
        mode = 'valid'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        ndim = 2
        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    x = xp.ones([1, 3], dtype=dtype)
                    y = xp.ones([1, 1], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x, y, ndim, mode=mode), backend.cpu_device)
                    npt.assert_allclose(W, [[1, 1, 1]], atol=1e-5)

                    x = xp.ones([1, 3], dtype=dtype)
                    y = xp.ones([1, 2], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x, y, ndim, mode=mode), backend.cpu_device)
                    npt.assert_allclose(W, [[2, 2]], atol=1e-5)

                    x = xp.ones([1, 1, 3], dtype=dtype)
                    y = xp.ones([2, 1, 1], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x,
                            y,
                            ndim,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(W, [[[1, 1, 1]],
                                            [[1, 1, 1]]], atol=1e-5)

    def test_convolve_adjoint_filter_full(self):
        mode = 'full'
        devices = [backend.cpu_device]
        if config.cupy_enabled:
            devices.append(backend.Device(0))

        ndim = 2
        for device in devices:
            xp = device.xp
            with device:
                for dtype in [np.float32, np.float64,
                              np.complex64, np.complex128]:
                    x = xp.ones([1, 3], dtype=dtype)
                    y = xp.ones([1, 5], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x, y, ndim, mode=mode), backend.cpu_device)
                    npt.assert_allclose(W, [[3, 3, 3]], atol=1e-5)

                    x = xp.ones([1, 3], dtype=dtype)
                    y = xp.ones([1, 4], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x, y, ndim, mode=mode), backend.cpu_device)
                    npt.assert_allclose(W, [[3, 3]], atol=1e-5)

                    x = xp.ones([1, 1, 3], dtype=dtype)
                    y = xp.ones([2, 1, 5], dtype=dtype)
                    W = backend.to_device(
                        conv.convolve_adjoint_filter(
                            x,
                            y,
                            ndim,
                            mode=mode,
                            output_multi_channel=True),
                        backend.cpu_device)
                    npt.assert_allclose(W, [[[3, 3, 3]],
                                            [[3, 3, 3]]], atol=1e-5)



# -*- coding: utf-8 -*-
"""MRI applications.
"""
import numpy as np
import sigpy as sp

from sigpy.mri import linop


def _estimate_weights(y, weights, coord):
    if weights is None and coord is None:
        with sp.get_device(y):
            weights = (sp.rss(y, axes=(0, )) > 0).astype(y.dtype)

    return weights


class SenseRecon(sp.app.LinearLeastSquares):
    r"""SENSE Reconstruction.

    Considers the problem

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 +
        \frac{\lambda}{2} \| x \|_2^2

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
            Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Pruessmann, K. P., Weiger, M., Scheidegger, M. B., & Boesiger, P.
        (1999).
        SENSE: sensitivity encoding for fast MRI.
        Magnetic resonance in medicine, 42(5), 952-962.

        Pruessmann, K. P., Weiger, M., Bornert, P., & Boesiger, P. (2001).
        Advances in sensitivity encoding with arbitrary k-space trajectories.
        Magnetic resonance in medicine, 46(4), 638-651.

    """

    def __init__(self, y, mps, lamda=0, weights=None,
                 coord=None, device=sp.cpu_device, coil_batch_size=None,
                 comm=None, show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        coil_batch_size=coil_batch_size, comm=comm)

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, lamda=lamda, show_pbar=show_pbar, **kwargs)


class SenseConstrainedRecon(sp.app.L2ConstrainedMinimization):
    r"""SENSE constrained reconstruction.

    Considers the problem

    .. math::
        \min_x &\| x \|_2^2 \\
        \text{s.t.} &\| P F S x - y \|_2^2 \le \epsilon

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        eps (float): constraint parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    See also:
       SenseRecon

    """

    def __init__(self, y, mps, eps,
                 weights=None, coord=None,
                 device=sp.cpu_device, coil_batch_size=None,
                 comm=None, show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        comm=comm, coil_batch_size=coil_batch_size)
        proxg = sp.prox.L2Reg(A.ishape, 1)
        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg, eps, show_pbar=show_pbar, **kwargs)


class L1WaveletRecon(sp.app.LinearLeastSquares):
    r"""L1 Wavelet regularized reconstruction.

    Considers the problem

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 + \lambda \| W x \|_1

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, W is the wavelet operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        wave_name (str): wavelet name.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Lustig, M., Donoho, D., & Pauly, J. M. (2007).
        Sparse MRI: The application of compressed sensing for rapid MR imaging.
        Magnetic Resonance in Medicine, 58(6), 1082-1195.

    """

    def __init__(self, y, mps, lamda,
                 weights=None, coord=None,
                 wave_name='db4', device=sp.cpu_device,
                 coil_batch_size=None, comm=None, show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        comm=comm, coil_batch_size=coil_batch_size)
        img_shape = mps.shape[1:]
        W = sp.linop.Wavelet(img_shape, wave_name=wave_name)
        proxg = sp.prox.UnitaryTransform(sp.prox.L1Reg(W.oshape, lamda), W)

        def g(input):
            device = sp.get_device(input)
            xp = device.xp
            with device:
                return lamda * xp.sum(xp.abs(W(input)))
        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg=proxg, g=g, show_pbar=show_pbar, **kwargs)


class L1WaveletConstrainedRecon(sp.app.L2ConstrainedMinimization):
    r"""L1 wavelet regularized constrained reconstruction.

    Considers the problem

    .. math::
        \min_x &\| W x \|_1 \\
        \text{s.t.} &\| P F S x - y \|_2^2 \le \epsilon

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, W is the wavelet operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        eps (float): constraint parameter.
        wave_name (str): wavelet name.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    See also:
       :func:`sigpy.mri.app.WaveletRecon`

    """

    def __init__(
            self, y, mps, eps,
            wave_name='db4', weights=None, coord=None,
            device=sp.cpu_device, coil_batch_size=None, comm=None,
            show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        comm=comm, coil_batch_size=coil_batch_size)
        img_shape = mps.shape[1:]
        W = sp.linop.Wavelet(img_shape, wave_name=wave_name)
        proxg = sp.prox.UnitaryTransform(sp.prox.L1Reg(W.oshape, 1), W)

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg, eps, show_pbar=show_pbar, **kwargs)


class TotalVariationRecon(sp.app.LinearLeastSquares):
    r"""Total variation regularized reconstruction.

    Considers the problem:

    .. math::
        \min_x \frac{1}{2} \| P F S x - y \|_2^2 + \lambda \| G x \|_1

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, G is the gradient operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        lamda (float): regularization parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    References:
        Block, K. T., Uecker, M., & Frahm, J. (2007).
        Undersampled radial MRI with multiple coils.
        Iterative image reconstruction using a total variation constraint.
        Magnetic Resonance in Medicine, 57(6), 1086-1098.

    """

    def __init__(self, y, mps, lamda,
                 weights=None, coord=None, device=sp.cpu_device,
                 coil_batch_size=None, comm=None, show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        comm=comm, coil_batch_size=coil_batch_size)

        G = sp.linop.Gradient(A.ishape)
        proxg = sp.prox.L1Reg(G.oshape, lamda)

        def g(x):
            device = sp.get_device(x)
            xp = device.xp
            with device:
                return lamda * xp.sum(xp.abs(x))

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg=proxg, g=g, G=G, show_pbar=show_pbar,
                         **kwargs)


class TotalVariationConstrainedRecon(sp.app.L2ConstrainedMinimization):
    r"""Total variation regularized constrained reconstruction.

    Considers the problem

    .. math::
        \min_x &\| G x \|_1 \\
        \text{s.t.} &\| P F S x - y \|_2^2 \le \epsilon

    where P is the sampling operator, F is the Fourier transform operator,
    S is the SENSE operator, G is the gradient operator,
    x is the image, and y is the k-space measurements.

    Args:
        y (array): k-space measurements.
        mps (array): sensitivity maps.
        eps (float): constraint parameter.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        device (Device): device to perform reconstruction.
        coil_batch_size (int): batch size to process coils.
        Only affects memory usage.
        comm (Communicator): communicator for distributed computing.
        **kwargs: Other optional arguments.

    See also:
       :func:`sigpy.mri.app.TotalVariationRecon`

    """

    def __init__(
            self, y, mps, eps,
            weights=None, coord=None, device=sp.cpu_device,
            coil_batch_size=None, comm=None, show_pbar=True, **kwargs):
        weights = _estimate_weights(y, weights, coord)
        if weights is not None:
            y = sp.to_device(y * weights**0.5, device=device)
        else:
            y = sp.to_device(y, device=device)

        A = linop.Sense(mps, coord=coord, weights=weights,
                        comm=comm, coil_batch_size=coil_batch_size)
        G = sp.linop.Gradient(A.ishape)
        proxg = sp.prox.L1Reg(G.oshape, 1)

        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        super().__init__(A, y, proxg, eps, G=G, show_pbar=show_pbar, **kwargs)


class JsenseRecon(sp.app.App):
    r"""JSENSE reconstruction.

    Considers the problem

    .. math::
        \min_{l, r} \frac{1}{2} \| l \ast r - y \|_2^2 +
        \frac{\lambda}{2} (\| l \|_2^2 + \| r \|_2^2)

    where \ast is the convolution operator.

    Args:
        y (array): k-space measurements.
        mps_ker_width (int): sensitivity maps kernel width.
        ksp_calib_width (int): k-space calibration width.
        lamda (float): regularization parameter.
        device (Device): device to perform reconstruction.
        weights (float or array): weights for data consistency.
        coord (None or array): coordinates.
        max_iter (int): Maximum number of iterations.
        max_inner_iter (int): Maximum number of inner iterations.

    References:
        Ying, L., & Sheng, J. (2007).
        Joint image reconstruction and sensitivity estimation in SENSE
        (JSENSE).
        Magnetic Resonance in Medicine, 57(6), 1196-1202.

        Uecker, M., Hohage, T., Block, K. T., & Frahm, J. (2008).
        Image reconstruction by regularized nonlinear inversion-
        joint estimation of coil sensitivities and image content.
        Magnetic Resonance in Medicine, 60(#), 674-682.

    """

    def __init__(self, y,
                 mps_ker_width=16, ksp_calib_width=24,
                 lamda=0, device=sp.cpu_device, comm=None,
                 weights=None, coord=None, max_iter=10,
                 max_inner_iter=10, show_pbar=True):
        self.y = y
        self.mps_ker_width = mps_ker_width
        self.ksp_calib_width = ksp_calib_width
        self.lamda = lamda
        self.weights = weights
        self.coord = coord
        self.max_iter = max_iter
        self.max_inner_iter = max_inner_iter

        self.device = sp.Device(device)
        self.comm = comm
        self.dtype = y.dtype
        self.num_coils = len(y)
        if comm is not None:
            show_pbar = show_pbar and comm.rank == 0

        self._get_data()
        self._get_vars()
        self._get_alg()
        super().__init__(self.alg, show_pbar=show_pbar)

    def _get_data(self):
        if self.coord is None:
            self.img_shape = list(self.y.shape[1:])
            ndim = len(self.img_shape)

            self.y = sp.resize(
                self.y, [self.num_coils] + ndim * [self.ksp_calib_width])

            if self.weights is not None:
                self.weights = sp.resize(
                    self.weights, ndim * [self.ksp_calib_width])

        else:
            self.img_shape = sp.estimate_shape(self.coord)
            calib_idx = np.amax(np.abs(self.coord), axis=-
                                1) < self.ksp_calib_width / 2

            self.coord = self.coord[calib_idx]
            self.y = self.y[:, calib_idx]

            if self.weights is not None:
                self.weights = self.weights[calib_idx]

        if self.weights is None:
            self.y = sp.to_device(self.y / np.abs(self.y).max(), self.device)
        else:
            self.y = sp.to_device(self.weights**0.5 *
                                  self.y / np.abs(self.y).max(), self.device)

        if self.coord is not None:
            self.coord = sp.to_device(self.coord, self.device)
        if self.weights is not None:
            self.weights = sp.to_device(self.weights, self.device)

        self.weights = _estimate_weights(self.y, self.weights, self.coord)

    def _get_vars(self):
        ndim = len(self.img_shape)

        mps_ker_shape = [self.num_coils] + [self.mps_ker_width] * ndim
        if self.coord is None:
            img_ker_shape = [i + self.mps_ker_width -
                             1 for i in self.y.shape[1:]]
        else:
            grd_shape = sp.estimate_shape(self.coord)
            img_ker_shape = [i + self.mps_ker_width - 1 for i in grd_shape]

        self.img_ker = sp.dirac(
            img_ker_shape, dtype=self.dtype, device=self.device)
        with self.device:
            self.mps_ker = self.device.xp.zeros(
                mps_ker_shape, dtype=self.dtype)

    def _get_alg(self):
        def min_mps_ker():
            self.A_mps_ker = linop.ConvImage(
                self.mps_ker.shape,
                self.img_ker,
                coord=self.coord,
                weights=self.weights)
            sp.app.LinearLeastSquares(
                self.A_mps_ker,
                self.y,
                self.mps_ker,
                lamda=self.lamda,
                max_iter=self.max_inner_iter,
                show_pbar=self.show_pbar).run()

        def min_img_ker():
            self.A_img_ker = linop.ConvSense(
                self.img_ker.shape,
                self.mps_ker,
                coord=self.coord,
                weights=self.weights,
                comm=self.comm)
            sp.app.LinearLeastSquares(
                self.A_img_ker,
                self.y,
                self.img_ker,
                lamda=self.lamda,
                max_iter=self.max_inner_iter,
                show_pbar=self.show_pbar).run()

        self.alg = sp.alg.AltMin(
            min_mps_ker, min_img_ker, max_iter=self.max_iter)

    def _output(self):
        xp = self.device.xp
        # Normalize by root-sum-of-squares.
        with self.device:
            mps_rss = 0
            mps = np.empty([self.num_coils] + self.img_shape, dtype=self.dtype)
            for c in range(self.num_coils):
                mps_c = sp.ifft(sp.resize(self.mps_ker[c], self.img_shape))
                mps_rss += xp.abs(mps_c)**2
                sp.copyto(mps[c], mps_c)

            mps_rss = sp.to_device(mps_rss)
            if self.comm is not None:
                self.comm.allreduce(mps_rss)

            mps_rss = mps_rss**0.5
            mps /= mps_rss
            return mps



# -*- coding: utf-8 -*-
"""MRI linear operators.

This module mainly contains the Sense linear operator,
which integrates multi-channel coil sensitivity maps and
discrete Fourier transform.

"""
import sigpy as sp


def Sense(mps, coord=None, weights=None, ishape=None,
          coil_batch_size=None, comm=None):
    """Sense linear operator.

    Args:
        mps (array): sensitivity maps of length = number of channels.
        coord (None or array): coordinates.
        weights (None or array): k-space weights.
        Useful for soft-gating or density compensation.
        ishape (None or tuple): image shape.
        coil_batch_size (None or int): batch size for processing multi-channel.
            When None, process all coils at the same time.
            Useful for saving memory.
        comm (None or `sigpy.Communicator`): communicator
            for distributed computing.

    """
    # Get image shape and dimension.
    num_coils = len(mps)
    if ishape is None:
        ishape = mps.shape[1:]
        img_ndim = mps.ndim - 1
    else:
        img_ndim = len(ishape)

    # Serialize linop if coil_batch_size is smaller than num_coils.
    num_coils = len(mps)
    if coil_batch_size is None:
        coil_batch_size = num_coils

    if coil_batch_size < len(mps):
        num_coil_batches = (num_coils + coil_batch_size - 1) // coil_batch_size
        A = sp.linop.Vstack([Sense(mps[c::num_coil_batches], coord=coord,
                                   weights=weights, ishape=ishape)
                             for c in range(num_coil_batches)], axis=0)

        if comm is not None:
            C = sp.linop.AllReduceAdjoint(ishape, comm, in_place=True)
            A = A * C

        return A

    # Create Sense linear operator
    S = sp.linop.Multiply(ishape, mps)
    if coord is None:
        F = sp.linop.FFT(S.oshape, axes=range(-img_ndim, 0))
    else:
        F = sp.linop.NUFFT(S.oshape, coord)

    A = F * S

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(F.oshape, weights**0.5)

        A = P * A

    if comm is not None:
        C = sp.linop.AllReduceAdjoint(ishape, comm, in_place=True)
        A = A * C

    A.repr_str = 'Sense'
    return A


def ConvSense(img_ker_shape, mps_ker, coord=None, weights=None, comm=None):
    """Convolution linear operator with sensitivity maps kernel in k-space.

    Args:
        img_ker_shape (tuple of ints): image kernel shape.
        mps_ker (array): sensitivity maps kernel.
        coord (array): coordinates.

    """
    ndim = len(img_ker_shape)
    A = sp.linop.ConvolveInput(
        img_ker_shape, mps_ker, mode='valid', output_multi_channel=True)

    if coord is not None:
        num_coils = mps_ker.shape[0]
        grd_shape = [num_coils] + sp.estimate_shape(coord)
        iF = sp.linop.IFFT(grd_shape, axes=range(-ndim, 0))
        N = sp.linop.NUFFT(grd_shape, coord)
        A = N * iF * A

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(A.oshape, weights**0.5)

        A = P * A

    if comm is not None:
        C = sp.linop.AllReduceAdjoint(img_ker_shape, comm, in_place=True)
        A = A * C

    return A


def ConvImage(mps_ker_shape, img_ker, coord=None, weights=None):
    """Convolution linear operator with image kernel in k-space.

    Args:
        mps_ker_shape (tuple of ints): sensitivity maps kernel shape.
        img_ker (array): image kernel.
        coord (array): coordinates.

    """
    ndim = img_ker.ndim

    A = sp.linop.ConvolveFilter(mps_ker_shape, img_ker,
                                mode='valid', output_multi_channel=True)

    if coord is not None:
        num_coils = mps_ker_shape[0]
        grd_shape = [num_coils] + sp.estimate_shape(coord)
        iF = sp.linop.IFFT(grd_shape, axes=range(-ndim, 0))
        N = sp.linop.NUFFT(grd_shape, coord)
        A = N * iF * A

    if weights is not None:
        with sp.get_device(weights):
            P = sp.linop.Multiply(A.oshape, weights**0.5)
        A = P * A

    return A
